"""
The Slit class for SONG.

Aims:
-----
store parameters for reduction

"""

import os

import joblib
import numpy as np
from astropy import table
from astropy.io import fits

from twodspec import thar
from twodspec.aperture import Aperture
from twodspec.ccd import CCD
from . import __path__
import matplotlib.pyplot as plt
import warnings
from ipyparallel import Client
from collections import OrderedDict

SLITINFO_TENERIFE = """
Slit width(μm) width(") Length(") Resolution(λ/Δλ) Sampling(pix) 
1    ø20,ø100  -        -         -                -
2    ø20       ø0.69    -         180000           1.35
3    100       3.44     10        36000            6.76
4    60        2.06     10        60000            4.05
5    45        1.55     10        80000            3.03           √
6    36        1.24     10        100              2.43
7    30        1.03     10        120000           2.02
8    25        0.86     10        145000           1.69
9    2mm hole  -        -         -                -
"""

SLITINFO_DELINGHA = """
Slit width(μm) width(") Length(") Resolution(λ/Δλ) Sampling(pix)
1    ø20,ø100  -        -         -                -
2    ø20       ø0.69    -         180000           1.35
3    100       3.44     10        36000            6.76
4    60        2.06     10        60000            4.05
5    45        1.55     10        80000            3.03          *
6    36        1.24     10        100000           2.43          *
7    30        1.03     10        120000           2.02          *
8    25        0.86     10        145000           1.69          *
9    20        0.69     10        181000           1.35
10   2mm hole  -        -         -                -
"""


class Slit:
    # io directory
    extdir = None  # None or a directory

    node = ""
    # slit number
    slit = 5

    # master frames
    # I don't think in this algorithm the bias and flat should be written to fits.
    # They can just stay in the Slit instance.
    fps_bias = []
    fps_flat = []
    bias = 0
    flat = 0
    bg = 0

    # blaze & sensitivity
    blaze = 0
    sensitivity = 0

    # apertures
    ap = None

    # wavelength solutions
    tws = table.Table([])

    # reduction parameters for SONG-China
    # 1. read images
    kwargs_read = dict(hdu=0, gain=1., ron=0., unit='adu', trim=None, rot90=0)
    # 2. trace orders
    kwargs_trace = dict(ap_width=15, sigma=7, maxdev=7, polydeg=4)
    # 3. scattered-light background for FLAT and STAR
    kwargs_background_flat = dict(q=(30, 1), kernel_size=(17, 17), sigma=(11, 7))
    kwargs_background_star = dict(q=(45, 45), kernel_size=(17, 17), sigma=(11, 7))
    # 4. normalize FLAT
    kwargs_normflat = dict(max_dqe=0.04, min_snr=20, smooth_blaze=5,
                           n_chunks=8, profile_oversample=10,
                           profile_smoothness=1e-2, num_sigma_clipping=20,
                           gain=1., ron=0, n_jobs=1)  # root directory
    # 5. fit grating equation
    kwargs_grating = dict(deg=(4, 10), nsigma=3, min_select=900)
    # 6. extract 1D spectrum
    kwargs_extract = dict(n_chunks=8, profile_oversample=10,
                          profile_smoothness=1e-2, num_sigma_clipping=20,
                          gain=1., ron=0, n_jobs=1)

    # SONG node list
    nodelist = ["delingha", "tenerife"]

    # thar template
    wave_temp = None
    thar_temp = None
    # load thar line list
    thar_line_list = None

    def set_parameters(self, node="delingha"):
        """ defaults to delingha parameters """
        assert node in self.nodelist
        if node == "delingha":
            # that's the default (2048x2048)
            pass
        if node == "tenerife":
            # for tenerife data (2088x2048)
            self.kwargs_read["trim"] = "[:, 21:2068]"

        # old comments
        # """ configuration for Song-China data """
        # # s.cfg.gain = {}
        # # s.cfg.trim['fits_section']=None
        # # s.cfg.rot90 = 0
        # """ configuration for Song-Tenerife data """
        # # s = Song._init_from_dir("/pool/song/sstenerife/star_spec/2017/20170314/night/raw/", verbose=False)
        # # s.cfg.trim['fits_section']="[:, 21:2068]" for tenerife data
        return

    def read_image(self, fp):
        """ read an image """
        if isinstance(fp, str):
            # single image
            return np.array(CCD.read(fp, **self.kwargs_read))
        else:
            # multiple images
            fp = list(fp)
            return np.array([self.read_image(_) for _ in fp])

    @staticmethod
    def read_header(fp):
        """ read an image header """
        return fits.getheader(fp)

    def proc_bias(self, fp):
        """ process bias """
        print("@Slit[{}]: processing bias ...".format(self.slit), end="")
        self.fps_bias = fp
        bias_data = self.read_image(fp)
        if bias_data.ndim == 3:
            self.bias = np.median(bias_data, axis=0)
        elif bias_data.ndim == 2:
            self.bias = bias_data
        print("Done!")
        return

    def proc_flat(self, fp):
        """ process bias """
        # 1. combine flat
        print("@Slit[{}]: processing flat ...".format(self.slit))
        self.fps_flat = fp
        flat_data = self.read_image(fp)
        if flat_data.ndim == 3:
            self.flat = np.median(flat_data, axis=0)
        elif flat_data.ndim == 2:
            self.flat = flat_data

        # 2. subtract bias
        flat_bias = self.flat - self.bias

        # 3. trace orders
        print("@Slit[{}]: tracing orders ...".format(self.slit))
        self.ap = Aperture.trace(flat_bias, method="naive", **self.kwargs_trace)

        # 3. subtract background
        print("@Slit[{}]: modeling background ...".format(self.slit))
        self.bg = self.ap.background(flat_bias, **self.kwargs_background_flat)
        flat_bias_bg = flat_bias - self.bg

        # 4. extract blaze & sensitivity
        print("@Slit[{}]: extracting blaze & sensitivity ...".format(self.slit), end="")
        self.blaze, self.sensitivity = self.ap.make_normflat(flat_bias_bg, **self.kwargs_normflat)
        print("Done!")
        return

    def proc_thar(self, fp, clear=True, ipcprofile=None):
        if isinstance(fp, list):
            # multiple thar
            print("@Slit[{}]: cleared tws ...".format(self.slit))
            if ipcprofile is None:
                print("@Slit[{}]: processing {} thar sequentially ...".format(self.slit, len(fp)))
                tws = [self.proc_thar(fp_, clear=True) for fp_ in fp]
                if clear:
                    self.clear_tws()
                for _ in tws:
                    self.tws.add_row(_)
                self.tws.sort("jdmid")
            else:
                # use ipcluster
                rc = Client(profile=ipcprofile)
                print("@Slit[{}]: dispatching {} thar to ipcluster (profile={}, nproc={}) ...".format(
                    self.slit, len(fp), ipcprofile, len(rc.ids)))
                dv = rc.direct_view()
                dv.block = True
                dv.push({"this_slit": self})
                dv.scatter("fp", fp)
                dv.execute("tws = [this_slit.proc_thar(_, clear=True) for _ in fp]")
                if clear:
                    self.clear_tws()
                # print(self.tws, dv.gather("tws"))
                # print(self.tws.colnames)
                for _ in dv.gather("tws"):
                    # print(_.keys())
                    self.tws.add_row(_)
                self.tws.sort("jdmid")
                dv.execute("%reset -f")
        else:
            # single thar

            # 1.read ThAr
            hdr = self.read_header(fp)
            # assert slit is correct
            assert hdr["SLIT"] == self.slit
            # thar data
            thar_data = self.read_image(fp)
            # thar time
            jdmid = hdr["JD-MID"]
            exptime = hdr["EXPTIME"]

            # 2.correct sensitivity
            thar_bg_sens = (thar_data - self.bias) / self.sensitivity
            # im_thar_denm_err = np.sqrt(np.abs(im_thar_denm)) / s.master_flats[slit]["norm"]

            # 3.extract ThAr spectrum
            rextr = self.ap.extract_all(thar_bg_sens, **self.kwargs_extract)
            thar_obs = rextr["spec_sum"]
            # thar_err = rextr["err_sum"]

            # 4.xcorrelate for initial wavelength guess
            wave_init = thar.corr_thar(self.wave_temp, self.thar_temp, thar_obs, maxshift=100)

            # 5. find thar lines
            tlines = thar.find_lines(wave_init, thar_obs, self.thar_line_list, npix_chunk=20)
            tlines = tlines[np.isfinite(tlines["line_x_ccf"])]

            # fit grating equation
            x = tlines["line_x_ccf"]  # line_x_ccf/line_x_gf
            y = tlines["order"]
            z = tlines["line"]
            pf1, pf2, indselect = thar.grating_equation(x, y, z, **self.kwargs_grating)
            tlines.add_column(table.Column(indselect, "indselect"))
            nlines = np.sum(indselect)
            # mpflux
            mpflux = np.median(tlines["line_peakflux"][tlines["indselect"]])
            # rms
            rms = np.std((pf2.predict(x, y) - z)[indselect])
            # predict wavelength solution
            nx, norder = thar_obs.shape
            mx, morder = np.meshgrid(np.arange(norder), np.arange(nx))
            wave_solu = pf2.predict(mx, morder)  # polynomial fitter

            # result
            calibration_dict = OrderedDict(
                fp=fp, jdmid=jdmid, exptime=exptime, wave_init=wave_init, wave_solu=wave_solu,
                tlines=tlines, nlines=nlines, rms=rms, pf1=pf1, pf2=pf2, mpflux=mpflux, thar_obs=thar_obs)
            # clear if necessary
            if clear:
                self.clear_tws()
            # add this thar to list
            if fp in self.tws["fp"]:
                # overwrite record
                idx = np.int(np.where(fp == self.tws["fp"])[0])
                self.tws.remove_row(idx)
            self.tws.add_row(calibration_dict)
            self.tws.sort("jdmid")
            return calibration_dict

    def proc_star(self, fp, write=True, ipcprofile=None, prefix="tstar", push=True):
        if isinstance(fp, list):
            # multiple star
            if len(fp) == 0:
                return []

            if ipcprofile is None:
                # use joblib
                print("@Slit[{}]: processing {} star sequentially ...".format(self.slit, len(fp)))
                results = []
                for fp_ in fp:
                    results.append(self.proc_star(fp_, write=True, prefix=prefix))
                # check invalid results
                for fp_ in results:
                    if prefix not in fp_:
                        warnings.warn(" >>> invalid data >>> {} ".format(fp_))
                return results

            else:
                # use ipcluster
                rc = Client(profile=ipcprofile)
                print("@Slit[{}]: dispatching {} files to ipcluster (profile={}, nproc={}) ...".format(
                    self.slit, len(fp), ipcprofile, len(rc.ids)))
                dv = rc.direct_view()
                dv.block = True
                if push:
                    dv.push({"this_slit": self, "prefix": prefix})
                dv.scatter("fp", fp)
                dv.execute("fps_out = this_slit.proc_star(fp, prefix=prefix)")
                # dv.execute()
                # dv.execute("tws = this_slit.tws")
                # self.tws = table.vstack(dv.gather("tws"))
                # self.tws.sort("jdmid")
                print("@Slit[{}]: Done!".format(self.slit))
                print("saved to files:")
                print("===========================================")
                results = dv.gather("fps_out")
                for fp_ in results:
                    print(fp_)
                print("===========================================")
                # dv.execute("%reset -f")
                # dv.execute("del()")
                # check invalid results
                for fp_ in results:
                    if prefix not in fp_:
                        warnings.warn(" >>> invalid data >>> {} ".format(fp_))
                return results
        else:
            # single star
            try:
                # 1.read star
                hdr = self.read_header(fp)
                # assert slit is correct
                assert hdr["SLIT"] == self.slit
                # star data
                star_data = self.read_image(fp)
                # star time
                jdmid = hdr["JD-MID"]
                exptime = hdr["EXPTIME"]
                bvc = hdr["BVC"]

                # 2.subtract bias & correct sensitivity
                star_bias_sens = (star_data - self.bias) / self.sensitivity

                # 3. subtract background
                bg = self.ap.background(star_bias_sens, **self.kwargs_background_star)
                star_bias_sens_bg = star_bias_sens - bg

                # 4.extract star spectrum
                rextr = self.ap.extract_all(star_bias_sens_bg, **self.kwargs_extract)
                # star_obs = rextr["spec_sum"]
                # star_err = rextr["err_sum"]

                # 5. append wavelength solution
                id_tws = np.argmin(np.abs(self.tws["jdmid"]-jdmid))
                rextr["wave"] = self.tws["wave_solu"][id_tws]
                rextr["wave_rms"] = self.tws["rms"][id_tws]
                rextr["blaze"] = self.blaze

                # 6. append info
                rextr["jdmid"] = jdmid
                rextr["exptime"] = exptime
                rextr["bvc"] = bvc

                # convert to table
                tstar = table.Table([rextr])
                tstar.meta = OrderedDict(hdr)

                # colname mapping
                tstar.rename_columns(['err_extr', 'err_extr1', 'err_extr2', 'err_sum', 'mask_extr',
                                      'spec_extr', 'spec_extr1', 'spec_extr2', 'spec_sum'],
                                     ['err', 'err1', 'err2', 'err_sum', 'mask',
                                      'flux', 'flux1', 'flux2', 'flux_sum'])

                if not write:
                    return tstar
                assert os.path.exists(self.extdir)
                fp_output = "{}/slit{}_{}_{}".format(self.extdir, self.slit, prefix, os.path.basename(fp))
                print("@Slit[{}]: saving to {} ...".format(self.slit, fp_output))
                tstar.write(fp_output, overwrite=True)
                return fp_output

            except Exception:
                return fp

    def __init__(self, slit=5, node="delingha", extdir="", ignore_warnings=True):
        self.slit = slit
        self.set_parameters(node=node)
        if extdir is not None:
            assert os.path.exists(extdir)
            self.extdir = extdir

        self.load_thar_template()
        self.clear_tws()

        # ignore warnings
        if ignore_warnings:
            self.ignore_warnings()
        return

    def __repr__(self):
        return "<Slit[{}] tws:{}>".format(self.slit, len(self.tws))

    def choose_tws(self, jd):
        pass

    def load_thar_template(self):
        # load thar template
        self.wave_temp = joblib.load("{}/calibration/wave_temp.dump".format(__path__[0]))
        self.thar_temp = joblib.load("{}/calibration/thar_temp.dump".format(__path__[0]))
        # thar_obs = joblib.load("{}/calibration/thar_obs.dump".format(__path__[0]))
        # load thar line list
        self.thar_line_list = joblib.load("{}/calibration/thar_line_list.dump".format(__path__[0]))
        return

    def clear_tws(self):
        """ clear ThAr Wavelength Solutions (tws) """
        self.tws = table.Table(
            data=[[], [], [], [], [], [], [], [], [], [], [], []],
            names=["fp", "jdmid", "exptime", "wave_init", "wave_solu", "tlines", "nlines", "rms", "pf1", "pf2", "mpflux", "thar_obs"],
            dtype=[object, object, object, object, object, object, object, object, object, object, object, object])
        return

    def ignore_warnings(self):
        warnings.filterwarnings("ignore")
        return

    def help(self):
        if self.node == "delingha":
            print(SLITINFO_DELINGHA)
        elif self.node == "tenerife":
            print(SLITINFO_TENERIFE)

    @staticmethod
    def proc_thar_static(slit, fp):
        # proc single thar
        print("@Slit[{}]: processing {} thar sequentially ...".format(slit.slit, len(fp)))
        return slit.proc_thar(fp, add=False)



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Fri Feb 24 16:00:00 2017

Modifications
-------------
- Fri Mar 27 21:34:55 2020

Aims
----
- Song class

Notes
-----
The SONG-China project defines file names with local time. i.e.,
>>> import time
>>> time.localtime()

"""

import glob
import os
from collections import OrderedDict
from datetime import datetime

import joblib
import numpy as np
from astropy.io import fits
from astropy.table import Table, Column
from astropy.time import Time
from joblib import Parallel, delayed
from matplotlib import pyplot as plt

from twodspec import extract, thar
from twodspec.aperture import Aperture
from twodspec.ccd import CCD
from .slit import Slit
from .utils import scan_files


class Song(Table):
    """ represent SONG configuration """

    # define all image types, used for classification
    ALL_IMGTYPE = [
        "BIAS", "FLAT", "FLATI2", "THAR", "THARI2", "STAR", "STARI2", "TEST"]

    # configuration for:
    kwargs_scan = dict(n_jobs=2, verbose=True)
    # 1. reading CCD images
    kwargs_read = dict(hdu=0, gain=1., ron=0., unit='adu', trim=None, rot90=0)
    # 2. combine images
    kwargs_combine = dict(method="median")
    # 3. scattered-light background for FLAT and STAR
    kwargs_background_flat = dict(q=(30, 1), kernel_size=(17, 17), sigma=(11, 7))
    kwargs_background_star = dict(q=(45, 45), kernel_size=(17, 17), sigma=(11, 7))
    # 4. normalize FLAT
    kwargs_normflat = dict(max_dqe=0.04, min_snr=20, smooth_blaze=5,
                           n_chunks=8, profile_oversample=10,
                           profile_smoothness=1e-2, num_sigma_clipping=20,
                           gain=1., ron=0, n_jobs=1)  # root directory
    # 5. extract 1D spectrum
    kwargs_extract = dict(n_chunks=8, profile_oversample=10,
                          profile_smoothness=1e-2, num_sigma_clipping=20,
                          gain=1., ron=0, n_jobs=1)

    # data & work directory
    rootdir = ""
    rawdir = ""
    extdir = ""
    rawdir1 = ""
    subdir = ""
    datejd0 = 0
    date = ""
    jdnight = True
    node = 2

    # unique slits
    unique_slits = []
    slits = []

    # master BIAS
    master_bias_fp = ""
    master_bias = None
    master_ron = None

    # master FLAT for each slit
    master_flats = dict()

    # THAR/STAR/FLATI2/TEST will be processed separately
    thar_list = dict()
    thari2_list = dict()

    star_list = list()
    stari2_list = list()
    flati2_list = list()
    test_list = list()

    # load ThAr template in Song class
    # print("@SONG: [ThAr] loading ThAr template ...")
    # thar_line_list = np.loadtxt(__path__[0] + "/calibration/thar.dat")
    # thar_solution_temp = (
    #     fits.getdata(__path__[0] + "/calibration/thar_template.fits", ext=1),
    #     fits.getdata(__path__[0] + "/calibration/thar_template.fits", ext=2),
    #     fits.getdata(__path__[0] + "/calibration/thar_template.fits", ext=3))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def new(self):
        """ initiate a new blank Song object """
        s = self.init(self.rootdir, self.date, self.jdnight, self.kwargs_scan["n_jobs"], self.kwargs_scan["verbose"], self.subdir, self.node)
        s.kwargs_read = self.kwargs_read
        s.kwargs_combine = self.kwargs_combine
        s.kwargs_background = self.kwargs_background
        return s

    @staticmethod
    def init(rootdir="/Users/cham/projects/song/star_spec", date="20191030", jdnight=True,
             n_jobs=-1, verbose=True, subdir="night", node=2):
        """ initiate from a directory path

        Parameters
        ----------
        rootdir: string
            root directory of data
        date: string
            optional
        jdnight: bool
            if True, use jd to define a night
        n_jobs:
            joblib.Parallel arguments
        verbose:
            joblib.Parallel arguments
        subdir:
            night / day
        node:
            1: Tenerife, 2: China

        Returns
        -------
        Song object
        """

        # directory path
        datejd0 = date2jd0(date)
        datedir = "{}/{}/{}".format(rootdir, date, subdir)
        rawdir = "{}/{}/{}/raw".format(rootdir, date, subdir)
        extdir = "{}/{}/{}/ext".format(rootdir, date, subdir)
        rawdir1 = "{}/{}/{}/raw".format(rootdir, dateplus1(date), subdir)

        # assert directories exist
        assert os.path.exists(datedir)  # date dir
        assert os.path.exists(rawdir)  # raw dir
        # if jdnight:
        #     assert os.path.exists(rawdir1)
        # if extdir does not exist, make it
        if not os.path.exists(extdir):
            print("@SONG: making directory [{}]..".format(extdir))
            os.mkdir(extdir)

        # define a night, glob files
        fps = glob.glob(rawdir + "/s{}_*.fits".format(node))
        if jdnight:
            fps.extend(glob.glob(rawdir1 + "/s{}_*.fits".format(node)))
        fps = np.array(fps)
        fps.sort()

        # extract jd
        if jdnight:
            jds = np.array([fp2jd(fp) for fp in fps])
            fps = fps[(jds >= datejd0) & (jds < datejd0 + 1)]

        if len(fps) == 0:
            return Song()

        # scan files
        print("@SONG: scanning files ...")
        s = Song(scan_files(fps, n_jobs=n_jobs, verbose=verbose))

        # add info
        s.rawdir = rawdir
        s.rawdir1 = rawdir1
        s.rootdir = rootdir
        s.extdir = extdir
        s.datejd0 = datejd0
        s.date = date
        s.jdnight = jdnight
        s.subdir = subdir
        s.node = node
        s.kwargs_scan["n_jobs"] = n_jobs
        s.kwargs_scan["verbose"] = verbose

        # get unique slits
        s.unique_slits = list(s.unique_config(cfgkeys=("SLIT")))

        return s

    def pipeline_bias(self, n_select=120, method_select='random'):
        """ process bias
        Parameters
        ----------
        n_select:
            number of frames used
        method_select:
            {"random", "top", "bottom", "all"}
        """
        self.master_bias, self.master_ron = self.ezmaster(
            {"IMAGETYP": "BIAS"}, n_select=n_select, method_select=method_select,
            method_combine='median', std=True)
        # write to disk
        self.master_bias_fp = "{}/masterbias_{}.fits".format(self.extdir, self.date)
        print("@SONG: master bias written to {}".format(self.master_bias_fp))
        self.master_bias.write(self.master_bias_fp, overwrite=True)
        return

    def pipeline_flat(self, slits="all", n_jobs_trace=1, verbose=10):

        if slits == "all":
            slits = self.unique_slits
        else:
            slits = [slits]

        for slit in slits:
            print()
            print("@SONG: processing FLAT for SLIT = ", slit)

            # 1.combine FLAT
            print("@SONG: combining FLAT ...")
            flat = self.ezmaster({"SLIT": slit, "IMAGETYP": "FLAT"},
                                 n_select=120, method_select='all',
                                 method_combine='median')
            # 2.substract bias
            print("@SONG: substracting BIAS ...")
            flat = flat - self.master_bias

            # 3.write to disk
            flat_fp = self.extdir + "/masterflat_{}_slit{}.fits".format(self.date, slit)
            print("@SONG: writing to {} ...".format(flat_fp))
            flat.write(flat_fp, overwrite=True)

            # 4. trace apertures
            ap = Aperture.trace(flat, method="naive", ap_width=15, sigma=7, maxdev=7, polydeg=4)

            # # figure of apertures
            # fig = plt.figure(figsize=(8, 8))
            # plt.imshow(np.log10(flat.data))
            # plt.plot(ap.ap_center.T, ap.x, 'w')
            # plt.xlabel("X coordinate")
            # plt.ylabel("Y coordinate")
            # plt.title("SLIT {}".format(slit))
            # fig.tight_layout()

            # # figure of horizontal slice
            # fig = plt.figure(figsize=(12, 6))
            # plt.plot(flat[1024])
            # plt.vlines(ap.ap_center[:, 1023], 0, 10000)
            # plt.xlabel("X coordinate")
            # plt.ylabel("Y coordinate")
            # plt.title("SLIT {} | Slice of row 1024".format(slit))
            # fig.tight_layout()

            # 5.background
            print("@SONG: computing background ...")
            bg = CCD(ap.background(flat, **self.kwargs_background_flat))
            bg_fp = self.extdir + "/masterflat_{}_slit{}_bg.fits".format(self.date, slit)
            bg.write(bg_fp, overwrite=True)
            flat = flat - bg

            # 6. calculate blaze functions and sensitivity
            print("@SONG: computing blaze function & sentitivity ...")
            # blz_bg, norm_bg = extract.make_normflat(np.array(flat), ap, **self.kwargs_normflat)
            blaze, sensitivity = ap.make_normflat(flat, **self.kwargs_normflat)

            blaze = CCD(blaze)
            blaze_fp = self.extdir + "/masterblaze_{}_slit{}.fits".format(self.date, slit)
            print("@SONG: writing blaze function to {}...".format(blaze_fp))
            blaze.write(blaze_fp, overwrite=True)

            sensitivity = CCD(sensitivity)
            sensitivity_fp = self.extdir + "/mastersens_{}_slit{}.fits".format(self.date, slit)
            print("@SONG: writing sentivity to {}...".format(sensitivity_fp))
            sensitivity.write(sensitivity_fp, overwrite=True)

            self.master_flats[slit] = dict(
                # slit
                slit=slit,  # slit number
                # flat
                flat=flat,  # flat-bias
                flat_fp=flat_fp,  # flat fp
                # apertures
                ap=ap,
                # background
                bg=bg,
                bg_fp=bg_fp,
                # blaze function
                blaze=blaze,
                blaze_fp=blaze_fp,
                # sentivity
                sensitivity=sensitivity,
                sensitivity_fp=sensitivity_fp,
            )
            print("@SONG: finishing processing SLIT {}".format(slit))

        return

    def pipeline_thar(self, poly_order=(5, 10)):

        # subtable of ThAr images
        ind_thar = self["IMAGETYP"] == "THAR"
        print("@SONG: in total {} ThAr images found!".format(np.sum(ind_thar)))
        sthar = self["fps", "SLIT", "FILE", "MJD-MID"][ind_thar]
        print(sthar)

        # for each ThAr image
        for i_thar in range(len(sthar)):

            # 0.check ThAr existence
            file = sthar["FILE"][i_thar]
            if file in list(self.thar_list.keys()):
                print("@SONG: skipping {} !".format(file))
                continue
            else:
                # process this ThAr image
                print("@SONG: processing {} ...".format(file))

            # 1.read ThAr
            slit = sthar["SLIT"][i_thar]
            mjdmid = sthar["MJD-MID"][i_thar]
            thar_fp = sthar["fps"][i_thar]
            thar_fn = os.path.basename(thar_fp)
            im_thar = self.read(thar_fp)
            im_meta = im_thar.meta
            im_thar = im_thar.data

            # 2.sensitivity correction (de-norm)
            im_thar_denm = (im_thar - self.master_bias) / self.master_flats[slit]["norm"]
            #im_thar_denm_err = np.sqrt(np.abs(im_thar_denm)) / self.master_flats[slit]["norm"]

            # 3.extract ThAr spectrum
            ap = self.master_flats[slit]["ap"]
            rextr = extract.extract_all(im_thar_denm, ap, **self.kwargs_extract)
            thar_sp = rextr["spec_sum"]
            thar_err = rextr["err_sum"]
            """ extract all apertures """

            thar1d_simple = np.copy(np.flipud(thar_sp))

            # 4.calibration
            calibration_results = thar.calibrate(
                thar1d_simple, self.thar_solution_temp, self.thar_line_list,
                poly_order=poly_order, slit=slit)
            sgrid_fitted_wave = calibration_results[0]
            wave_final = np.copy(np.flipud(sgrid_fitted_wave))

            # save figures
            fig = calibration_results[2]
            fig.savefig(self.extdir + "thar{}_{:s}".format(
                slit, thar_fn.replace(".fits", "_diagnostics.pdf")))
            plt.close(fig)
            fig = calibration_results[3]
            fig.savefig(self.extdir + "thar{}_{:s}".format(
                slit, thar_fn.replace(".fits", "_used_lines.pdf")))
            plt.close(fig)

            # 5.ThAr results: [wave] [sp] [sp_err]
            exheader = OrderedDict([
                ("---PC---", "---PROCESSING---"),
                ("author", "Bo Zhang"),
                ("court", "home"),])
            thar_header_ = OrderedDict([
                ("--HDUS--", "---HDUs---"),
                ("HDU0", "wavelength solution"),
                ("HDU1", "1D ThAr spectrum"),
                ("HDU2", "1D ThAr spectrum error")])
            im_meta.update(exheader)
            im_meta.update(thar_header_)
            h0 = fits.hdu.PrimaryHDU(wave_final, fits.Header(im_meta))  # ThAr wavelength
            # h1 = fits.hdu.ImageHDU(wave_final)  # ThAr wavelength
            h1 = fits.hdu.ImageHDU(thar_sp)  # ThAr spectrum
            h2 = fits.hdu.ImageHDU(thar_err)  # ThAr error spectrum
            hl = fits.HDUList([h0, h1, h2])
            hl.writeto(self.extdir + "thar{}_{:s}".format(slit, thar_fn),
                       overwrite=True)
            # figure(); plot(wave_final.T, thar_sp.T)

            # record ThAr to self.thar_list
            self.thar_list[thar_fn] = dict(filename=thar_fn,
                                           mjdmid=mjdmid,
                                           slit=slit,
                                           wave_final=wave_final,
                                           thar1d_simple=thar1d_simple)

    def pipeline_star(self, key="STAR", n_jobs=-1, verbose=10):

        # subtable of ThAr images
        ind_star = self["IMAGETYP"] == key
        sub_star = np.where(ind_star)[0]
        print("@SONG: in total {} {} images found!".format(np.sum(ind_star), key))
        sstar = self["fps", "SLIT", "FILE", "MJD-MID"][ind_star]
        print(sstar)

        # prepare calibration data
        if key == "STAR":
            star_list = self.star_list
        elif key == "STARI2":
            star_list = self.stari2_list
        elif key == "FLATI2":
            star_list = self.flati2_list
        else:
            raise(AssertionError("@SONG: key *{}* is not valid!".format(key)))

        # skip processed files
        ind_unprocessed = np.zeros((len(sstar),), dtype=bool)
        for i_star in range(len(sstar)):
            # check STAR existence
            file = sstar["FILE"][i_star]
            if file not in list(star_list):
                # process this STAR image
                ind_unprocessed[i_star] = True
        print("==========================================================")
        print("@SONG: {} files will be skipped!".format(
            np.sum(np.logical_not(ind_unprocessed))))
        print("==========================================================")
        print(sstar[np.logical_not(ind_unprocessed)])
        print("==========================================================")
        print("")
        print("")
        print("==========================================================")
        print("@SONG: {} files will be processed!".format(
            np.sum(ind_unprocessed)))
        print("==========================================================")
        print(sstar[ind_unprocessed])
        print("==========================================================")
        print("")
        print("")
        sstar = Table(sstar[ind_unprocessed])

        r = Parallel(n_jobs=n_jobs, verbose=verbose)(
            delayed(self._pipeline_a_star)(self["fps"][i_star])
            for i_star in sub_star[ind_unprocessed])

        if key == "STAR":
            self.star_list.extend(r)
        elif key == "STARI2":
            self.stari2_list.extend(r)
        elif key == "FLATI2":
            self.flati2_list.extend(r)

        # # record results
        # for i_star in range(len(sstar)):
        #     star_list[sstar["FILE"][i_star]] = r[i_star]

        # if key == "STAR":
        #     self.star_list = star_list
        # elif key == "STARI2":
        #     self.stari2_list = star_list
        # elif key == "FLATI2":
        #     self.flati2_list = star_list

        return

    def select(self, cond_dict=None, method="all", n_select=10,
               returns=("fps"), verbose=False):
        """ select some images from list

        Parameters
        ----------
        cond_dict: dict
            the dict of colname:value pairs
        method: string, {"all", "random", "top", "bottom"}
            the method adopted
        n_select:
            the number of images that will be selected
            if n_images is larger than the number of images matched conditions,
            then n_images is forced to be n_matched
        returns:
            the column name(s) of the column that will be returned
            if returns == 'sub', return the subs of selected images
        verbose:
            if True, print result

        Returns
        -------
        the Song instance

        Examples
        --------
        >>> s.list_image({"IMAGETYP":"STAR"}, returns=["OBJECT"])
        >>> s.select({"IMAGETYP":"THAR", "SLIT":6}, method="all", n_select=200,
        >>>          returns="sub", verbose=False)

        """

        # determine the matched images
        ind_match = np.ones((len(self),), dtype=bool)
        if cond_dict is None or len(cond_dict) < 1:
            print("@SONG: no condition is specified!")
        for k, v in cond_dict.items():
            ind_match = np.logical_and(ind_match, self[k] == v)

        # if no image found
        n_matched = np.sum(ind_match)
        if n_matched < 1:
            # print("@SONG: no images matched!")
            return []

        sub_match = np.where(ind_match)[0]
        # determine the number of images to select
        n_return = np.min([n_matched, n_select])

        if verbose:
            print("@SONG: conditions are ", cond_dict)
            print("@SONG: {0} matched & {1} selected & {2} will be returned"
                  "".format(n_matched, n_select, n_return))

        # select according to method
        assert method in {"all", "random", "top", "bottom"}
        sub_rand = np.arange(0, n_matched, dtype=int)
        if method is "all":
            n_return = n_matched
        elif method is "random":
            np.random.shuffle(sub_rand)
            sub_rand = sub_rand[:n_return]
        elif method is "top":
            sub_rand = sub_rand[:n_return]
        elif method is "bottom":
            sub_rand = sub_rand[-n_return:]
        sub_return = sub_match[sub_rand]

        # constructing result to be returned
        if returns is "sub":
            result = sub_return
        else:
            result = self[returns][sub_return]

        # verbose
        if verbose:
            print("@SONG: these are all images selected")
            # here result is a Table
            print(result.__repr__())

        return result

    # #################################### #
    # simplified methods to select subsets
    # currently, just use select() method
    # #################################### #

    def ezselect_rand(self, cond_dict, n_select=10, returns="sub",
                      verbose=False):
        return self.select(cond_dict=cond_dict, returns=returns,
                           method="random", n_select=n_select, verbose=verbose)

    def ezselect_all(self, cond_dict, n_select=10, returns="sub",
                     verbose=False):
        return self.select(cond_dict=cond_dict, returns=returns,
                           method="all", n_select=n_select, verbose=verbose)

    # TODO: this method will either be updated/deleted
    def list_image(self, imagetp="FLAT", kwds=None, max_print=None):
        list_image(self, imagetp=imagetp, return_col=None, kwds=kwds,
                   max_print=max_print)
        return

    # #################################### #
    # methods to summarize data
    # #################################### #
    def unique_config(self, cfgkeys=("SLIT", "IMAGETYP")):
        """ find number of unique config """
        result = np.asarray(np.unique(self[cfgkeys]))
        print("@SONG: {0} unique config found!".format(len(result)), list(result))
        self.describe().pprint()
        return result

    def describe(self, cfgkeys=("SLIT", "IMAGETYP")):
        """ generate a table of stats on image configs

        Parameters
        ----------
        cfgkeys: tuple
            a pair of keys, default is ("SLIT", "IMAGETYP")

        Returns
        -------
        summary in Table format

        """
        # initialize result Table
        col0 = [Column(np.unique(self[cfgkeys[0]]), cfgkeys[0])]
        cols = [Column(np.zeros_like(col0[0], dtype=int), key2val) for key2val
                in np.unique(self[cfgkeys[1]])]
        col0.extend(cols)
        result = Table(col0)

        # do statistics & assign to result Table
        unique_result = np.unique(self[cfgkeys], return_counts=True)
        for keyvals_unique, count in zip(*unique_result):
            result[keyvals_unique[1]][
                result[cfgkeys[0]] == keyvals_unique[0]] = count

        return result

    @property
    def summary(self, colname_imagetype="IMAGETYP", return_data=False):
        """

        Parameters
        ----------
        colname_imagetype: string
            the keyword name for image type, default is "IMAGETYP"
        return_data: bool
            if True, return the summary data

        Returns
        -------
        unique images

        """

        u, uind, uinv, ucts = np.unique(self[colname_imagetype],
                                        return_counts=True, return_index=True,
                                        return_inverse=True)
        # print summary information
        print("=====================================================")
        print("[SUMMARY] {:s}".format(self.extdir))
        print("=====================================================")
        for i in range(len(u)):
            print("{:10s} {:d}".format(u[i], ucts[i]))
        print("=====================================================")
        self.describe().pprint()
        print("=====================================================")

        # return results
        if return_data:
            return u, uind, uinv, ucts

    def read(self, fp):
        """ read a CCD frame """
        return CCD.read(fp, **self.kwargs_read)

    def reads(self, fps, method="median", std=False):
        """ actually a combine method """
        return CCD.reads(fps, method=method, std=std, **self.kwargs_read)

    def ezmaster(self, cond_dict={"IMAGETYP": "BIAS"}, n_select=10,
                 method_select="top", method_combine="median", std=False):
        """

        Parameters
        ----------
        cond_dict: dict
            conditions on {"BIAS", "FLAT", "FLATI2", "THAR", "THARI2", "STAR", "STARI2", "TEST"}
        n_select: int
            number of images will be selected
        method_select:
            scheme of selection
        method_combine:
            method of combining
        std:
            if True, return average and std

        Returns
        -------
        combined image

        """

        assert method_select in {"random", "top", "bottom", "all"}

        # if any cond_dict key does not exist in song
        try:
            for k in cond_dict.keys():
                assert k in self.colnames
        except Exception:
            raise ValueError("@SONG: key not found: {0}".format(k))

        # find fps of matched images
        fps = self.select(cond_dict, method=method_select, n_select=n_select)
        print("@SONG: *ezmaster* working on ", fps)

        # combine all selected images
        return self.reads(fps, method=method_combine, std=std)

    def draw(self, save=None, figsize=(15, 6), return_fig=False):
        """ a description of observation """
        # ljd = self["JD-MID"] + 8 / 24
        # lhour = (np.mod(ljd, 1) + 0.5) * 24

        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)

        # sub table for text
        tsub = self["JD-MID", "SLIT", "IMAGETYP", "OBJECT"]
        # tsub["JD-MID"] = lhour
        tsub.sort("JD-MID")
        trep = find_all_repeats(tsub)

        _ms = 15
        ind = self["IMAGETYP"] == "BIAS"
        ax.plot(tsub["JD-MID"][ind], tsub["SLIT"][ind], "o", mfc="none", mec='g', ms=_ms, label="BIAS")
        ind = self["IMAGETYP"] == "FLAT"
        ax.plot(tsub["JD-MID"][ind], tsub["SLIT"][ind], "^", mfc="none", mec='r', ms=_ms, label="FLAT")
        ind = self["IMAGETYP"] == "FLATI2"
        ax.plot(tsub["JD-MID"][ind], tsub["SLIT"][ind], "^", mfc="none", mec='b', ms=_ms, label="FLATI2")
        ind = self["IMAGETYP"] == "THAR"
        ax.plot(tsub["JD-MID"][ind], tsub["SLIT"][ind], "D", mfc="none", mec='c', ms=_ms, mew=2, label="THAR")
        ind = self["IMAGETYP"] == "STAR"
        ax.plot(tsub["JD-MID"][ind], tsub["SLIT"][ind], "s", mfc="none", mec='m', ms=_ms, label="STAR")
        ind = self["IMAGETYP"] == "STARI2"
        ax.plot(tsub["JD-MID"][ind], tsub["SLIT"][ind], "s", mfc="none", mec='y', ms=_ms, label="STARI2")
        ind = self["IMAGETYP"] == "TEST"
        ax.plot(tsub["JD-MID"][ind], tsub["SLIT"][ind], "s", mfc="none", mec='gray', ms=_ms, label="TEST")

        # annotate the repeats
        for i in range(len(trep)):
            if trep['label'][i] not in ["BIAS", "FLAT", "FLATI2", "THAR", "THARI2"]:
                fontcolor = "r"
            else:
                fontcolor = (0, 0, 0, .8)

            if np.mod(i, 2) == 0:
                tick_height = 0.3
                text_kwargs = dict(horizontalalignment='left',
                                   verticalalignment='bottom', rotation=50,
                                   fontsize=12, color=fontcolor)
            else:
                tick_height = -0.3
                text_kwargs = dict(horizontalalignment='left',
                                   verticalalignment='top', rotation=-50,
                                   fontsize=12, color=fontcolor)

            if trep['c'][i] > 1:
                tick_x = [trep["jd1"][i], trep["jd1"][i], trep["jd2"][i], trep["jd2"][i]]
                tick_y = [trep["slit"][i], trep["slit"][i] + tick_height, trep["slit"][i] + tick_height, trep["slit"][i]]
                ax.plot(tick_x, tick_y, 'k')
                ax.text(trep["jd1"][i] * .5 + trep["jd2"][i] * .5, trep["slit"][i] + tick_height,
                        trep['label'][i] + " x {}".format(trep['c'][i]), **text_kwargs)
            else:
                tick_x = [trep["jd1"][i], trep["jd1"][i]]
                tick_y = [trep["slit"][i], trep["slit"][i] + tick_height]
                ax.plot(tick_x, tick_y, 'k')
                ax.text(trep["jd1"][i], trep["slit"][i] + tick_height,
                        trep['label'][i] + " x {}".format(trep['c'][i]), **text_kwargs)

        ax.legend(loc=3)
        ax.set_xlabel("Julian Day")
        ax.set_ylabel("Slit Number")
        ax.set_yticks(np.arange(9) + 1)
        _ylim = self["SLIT"].min() - 2, self["SLIT"].max() + 2
        ax.set_ylim(_ylim)
        ax.set_xticks(np.linspace(self.datejd0 - 8 / 24, self.datejd0 - 8 / 24 + 1, 13))
        ax.set_xlim(self.datejd0 - 8 / 24, self.datejd0 - 8 / 24 + 1)
        ax.grid(True, axis="y", ydata=np.arange(11))

        ax2 = ax.twiny()
        ax2.set_xlabel("{}-{}-{} Beijing local time (UTC+8 Hour)".format(self.date[:4], self.date[4:6], self.date[6:8]))
        ax2.set_xticks(np.linspace(12, 36, 13))
        ax2.set_xticklabels(["{:.0f}".format(_) for _ in np.mod(np.linspace(12, 36, 13), 24)])
        ax2.set_xlim(12, 36)
        ax2.grid(True, axis='x', linestyle='--')

        fig.tight_layout()

        if save is True:
            # save to default
            figpath = self.extdir+"/summary_{}.pdf".format(self.date)
            print("@SONG: saving summary figure to {}".format(figpath))
            fig.savefig(figpath)
        elif save is not None:
            figpath = save
            print("@SONG: saving summary figure to {}".format(figpath))
            fig.savefig(save)

        if return_fig:
            return fig
        else:
            plt.close(fig)
            return

    def _pipeline_a_star(self, star_fp):

        # read STAR
        hdr = fits.getheader(star_fp)
        slit = hdr["SLIT"]
        mjdmid = hdr["MJD-MID"]

        star_fn = os.path.basename(star_fp)

        n_ap = self.master_flats[slit]["ap"].n_ap
        im_star = self.read(star_fp)
        im_meta = im_star.meta
        im_star = im_star.data

        # find ThAr
        thar_keys = list(self.thar_list.keys())
        # check slit
        ind_good = np.zeros_like(thar_keys, dtype=bool)
        for i in range(len(thar_keys)):
            if self.thar_list[thar_keys[i]]["slit"] == slit:
                ind_good[i] = True
        thar_keys = np.array(thar_keys)[ind_good]

        thar_mjdmid = np.array([self.thar_list[_]["mjdmid"] for _ in thar_keys])
        # ThAr before STAR
        sub_bfr = np.where(mjdmid > thar_mjdmid)[0]
        if len(sub_bfr)>0:
            ind_bfr = np.argmax(thar_mjdmid[sub_bfr])
            ind_tharbfr = sub_bfr[ind_bfr]
            print("@SONG: mjdmid: ", mjdmid)
            for i in range(len(thar_keys)):
                print("@SONG: thar: ", thar_keys[i],
                      " mjd: ", self.thar_list[thar_keys[i]]["mjdmid"],
                      " slit: ", self.thar_list[thar_keys[i]]["slit"])
            tharbfr = self.thar_list[thar_keys[ind_tharbfr]]["wave_final"]
            tharbfr_file = thar_keys[ind_tharbfr]
            print("@SONG: STAR[{}] THAR0[{}]".format(star_fn, tharbfr_file))
        else:
            tharbfr = np.zeros((n_ap, 2048), dtype=float)
            tharbfr_file = ""
            print("@SONG: Couldn't find ThAr before observation!")
        # ThAr after STAR
        sub_aft = np.where(mjdmid < thar_mjdmid)[0]
        if len(sub_aft) == 0:
            tharaft = np.zeros_like(tharbfr)
            tharaft_file = ""
            print("@SONG: Couldn't find ThAr after observation!")
        else:
            ind_aft = np.argmin(thar_mjdmid[sub_aft])
            ind_tharaft = sub_aft[ind_aft]
            tharaft = self.thar_list[thar_keys[ind_tharaft]]["wave_final"]
            tharaft_file = thar_keys[ind_tharaft]
            print("@SONG: STAR[{}] THAR1[{}]".format(star_fn, tharaft_file))

        # sensitivity correction
        im_star_denm = (im_star - self.master_bias) / self.master_flats[slit]['norm']
        # im_star_denm_err = np.sqrt(np.abs(im_star_denm)) / flats[slit]['norm']

        # scattered light substraction
        ap = self.master_flats[slit]['ap']
        bg = ap.background(im_star_denm, **self.kwargs_background_star)

        # extract STAR
        rextr = extract.extract_all(im_star_denm-bg, self.master_flats[slit]["ap"],
                                    **self.kwargs_extract)
        star_sp = rextr["spec_sum"]
        star_err = rextr["err_sum"]

        star_optsp = rextr["spec_extr1"]
        star_opterr = rextr["err_extr1"]

        star_robsp = rextr["spec_extr2"]
        star_roberr = rextr["err_extr2"]

        star_mask = rextr["mask_extr"]

        """ STAR results: [wave] [sp] [sp_err]"""
        exheader = OrderedDict([
            ("---PC---", "---PROCESSING---"),
            ("author", "Bo Zhang"),
            ("court", "home"),])
        star_header_ = OrderedDict([
            ("-LAYERS-", "-LAYERS-"),
            ("layer0", "optimal extracted spectrum"),
            ("layer1", "simple extracted spectrum"),
            ("layer2", "blaze function"),
            ("layer3", "wavelength before observation"),
            ("layer4", "wavelength after observation"),
            ("layer5", "error of optimal extracted spectrum"),
            ("layer6", "error of simple extracted spectrum"),
            ("layer7", "robust extracted spectrum"),
            ("layer8", "error of robust extracted spectrum"),
            ("layer9", "mask of extracted spectrum"),
            ("tharbfr", tharbfr_file),
            ("tharaft", tharaft_file)])

        im_meta.update(exheader)
        im_meta.update(star_header_)
        data = np.array([star_optsp,  # opt spec
                         star_sp,  # simple spec
                         self.master_flats[slit]['blz'],  # blaze
                         tharbfr,  # wave before
                         tharaft,  # wave after
                         star_opterr,  # opt error
                         star_err,
                         star_robsp,
                         star_roberr,
                         star_mask])  # simple error

        h0 = fits.hdu.PrimaryHDU(data, fits.Header(im_meta))
        hl = fits.HDUList([h0])
        hl.writeto(self.extdir + "pstar{}_{:s}".format(slit, star_fn),
                   overwrite=True)

        return star_fn

    def daily(self, proc_slits="all", sharebias=True, star=True, stari2=False, flati2=False,
              ipcprofile="default"):
        """ daily pipeline """
        joblib.dump(self, "{}/{}_song.dump".format(self.extdir, self.date))
        print("[{}] starting daily pipeline ...".format(datetime.now()))
        print("===========================================")
        print("@Song: unique slits are ", self.unique_slits)
        print("===========================================")

        if proc_slits is "all":
            # proc all slits
            proc_slits = list(self.unique_slits)
        else:
            # proc the specified slits
            assert isinstance(proc_slits, list)
            for _ in proc_slits:
                assert _ in self.unique_slits

        slits = []
        for slit in proc_slits:
            this_slit = Slit(slit=slit, extdir=self.extdir)
            # ind
            if sharebias:
                # share bias
                ind_bias = self.ezselect_all({"IMAGETYP": "BIAS"})
                if len(ind_bias) > 120:
                    ind_bias = np.random.choice(ind_bias, 120)
            else:
                # don't share
                ind_bias = self.ezselect_all({"IMAGETYP": "BIAS", "SLIT": slit})
            ind_flat = self.ezselect_all({"IMAGETYP": "FLAT", "SLIT": slit})
            ind_thar = self.ezselect_all({"IMAGETYP": "THAR", "SLIT": slit})

            if any(_ is None for _ in [ind_bias, ind_flat, ind_thar]):
                print("@Song: skipping slit[{}] due to lack of calibration data ...".format(slit))

            ind_star = self.ezselect_all({"IMAGETYP": "STAR", "SLIT": slit})
            ind_flati2 = self.ezselect_all({"IMAGETYP": "FLATI2", "SLIT": slit})
            ind_stari2 = self.ezselect_all({"IMAGETYP": "STARI2", "SLIT": slit})

            sinfo = "@SONG-slit[{}]: will process ".format(this_slit.slit)
            if star:
                sinfo += " star [{}]".format(len(ind_star))
            if flati2:
                sinfo += " flati2 [{}]".format(len(ind_flati2))
            if stari2:
                sinfo += " stari2 [{}]".format(len(ind_stari2))
            print(sinfo)

            if ind_bias is None or ind_flat is None or ind_thar is None:
                # need calibration images!
                continue
            # bias
            fps_bias = self["fps"][ind_bias]
            this_slit.proc_bias(fps_bias)
            # flat
            fps_flat = self["fps"][ind_flat]
            this_slit.proc_flat(fps_flat)
            # thar
            fps_thar = list(self["fps"][ind_thar])
            this_slit.proc_thar(fps_thar, ipcprofile=ipcprofile)
            # star
            if star and ind_star is not None:
                fps_star = list(self["fps"][ind_star])
                this_slit.proc_star(fps_star, ipcprofile=ipcprofile, prefix="tstar")
            # flati2
            if flati2 and ind_flati2 is not None:
                fps_flati2 = list(self["fps"][ind_flati2])
                this_slit.proc_star(fps_flati2, ipcprofile=ipcprofile, prefix="tflati2")
            # stari2
            if stari2 and ind_stari2 is not None:
                fps_stari2 = list(self["fps"][ind_stari2])
                this_slit.proc_star(fps_stari2, ipcprofile=ipcprofile, prefix="tstari2")
            joblib.dump(this_slit, "{}/{}_slit{}.dump".format(self.extdir, self.date, slit))
            slits.append(this_slit)
            print("===========================================")
        print("DONE!~")
        print("===========================================")
        self.slits = slits
        return


# def _try_trace_apertures(flat, sigma_):
#     try:
#         ap = Aperture.trace(flat, method="canny", sigma=sigma_, verbose=False)
#         return ap.n_ap
#     except Exception as e_:
#         return -1


# used in draw()
def find_repeats(tsub, i1=0):
    c = 0
    for i in range(i1, len(tsub)):
        if tsub["SLIT", "IMAGETYP", "OBJECT"][i] == \
                tsub["SLIT", "IMAGETYP", "OBJECT"][i1]:
            c += 1
        else:
            break
    if tsub["IMAGETYP"][i1] in ["BIAS", "FLAT", "FLATI2", "THAR", "THARI2"]:
        this_label = tsub["IMAGETYP"][i1]
    else:
        this_label = tsub["OBJECT"][i1]
    this_mjd1 = tsub["JD-MID"][i1]
    this_mjd2 = tsub["JD-MID"][i1 + c - 1]
    this_slit = tsub["SLIT"][i1]
    return this_mjd1, this_mjd2, this_slit, this_label, c


def find_all_repeats(tsub):
    repeats = []
    i1 = 0
    while i1 < len(tsub):
        repeats.append(find_repeats(tsub, i1))
        i1 += repeats[-1][-1]
    return Table(np.array(repeats),
                 names=["jd1", "jd2", "slit", "label", "c"],
                 dtype=[float, float, int, str, int])


# random choice
def random_ind(n, m):
    """ from n choose m randomly """
    return np.argsort(np.random.rand(n,))[:m]


# list images of a specified type
def list_image(t, imagetp="FLAT", return_col=None, kwds=None, max_print=None):
    """ list images with specified IMAGETYP value

    Examples
    --------
    >>> list_image(t2, imagetp="STAR", kwds=["OBJECT"])

    Parameters
    ----------
    t: Table
        catalog of files, generated by *scan_files*
    imagetp: string
        IMAGETYP value
    kwds: list
        optional. additional columns to be displayed
    max_print:
        max line number

    Returns
    -------

    """
    ind_mst = np.where(t["IMAGETYP"] == imagetp)[0]

    if max_print is not None:
        if max_print > len(ind_mst):
            max_print = len(ind_mst)
    else:
        max_print = len(ind_mst)

    print("@SONG: these are all images of type %s" % imagetp)
    print("+ --------------------------------------------------")
    if isinstance(kwds, str):
        kwds = [kwds]
    if kwds is None or kwds == "":
        for i in range(max_print):
            print("+ %04d - %s" % (i, t["fps"][ind_mst[i]]))
    else:
        assert isinstance(kwds, list) or isinstance(kwds, tuple)
        for kwd in kwds:
            try:
                assert kwd in t.colnames
            except AssertionError:
                print("kwd", kwd)
                raise AssertionError()

        for i in range(max_print):
            s = "+ %04d - %s" % (i, t["fps"][ind_mst[i]])
            for kwd in kwds:
                s += "  %s" % t[kwd][ind_mst[i]]
            print(s)
    print("+ --------------------------------------------------")

    if return_col is not None:
        return t[return_col][ind_mst]


# CCD operations
def read_image(fp, hdu=0, gain=1., ron=0., unit='adu', trim=None, rot90=0):
    """ read image """
    ccds = CCD.read(fp, hdu=hdu, gain=gain, ron=ron, unit=unit, trim=trim, rot90=rot90)
    return ccds


# combine CCDs
def combine_images(fps, method="median", std=False, hdu=0, gain=1., ron=0., unit='adu', trim=None, rot90=0, **kwargs):
    """ combine images

    Parameters
    ----------
    fps:
        file paths
    method:
        mean or median
    std:
        False if you don't want std
        True if you want std, this is useful to estimate Read-Out-Noise
    kwargs:
        CCD.read() keyword args
    """
    ccds = [CCD.read(fp, hdu=hdu, gain=gain, ron=ron, unit=unit, trim=trim,
                     rot90=rot90) for fp in fps]

    if method == "mean":
        ccd_comb = CCD.mean(ccds)
    elif method == "median":
        ccd_comb = CCD.median(ccds)
    else:
        raise ValueError("@SONG: bad value for method [{}]".format(method))

    if not std:
        return ccd_comb
    else:
        ccd_std = CCD.std(ccds)
        return ccd_comb, ccd_std


# convert date string
# print(date2jd0("20191031"))
def date2jd0(s="20191031"):
    ds = "{}-{}-{}T12:00:00".format(s[0:4],s[4:6],s[6:8])
    return Time(ds,format="isot").jd


# print(jd02date(2458788.0+1))
def jd02date(jd):
    ds = Time(jd,format="jd").isot
    return ds[:10].replace("-", "")


# print(dateplus1("20191031"))
def dateplus1(s="20191031"):
    return jd02date(date2jd0(s)+1)


def fp2jd(fp="s2_2019-10-30T19-56-00.fits"):
    fn = os.path.basename(fp)
    dtisot = "{}:{}:{}".format(fn[3:16], fn[17:19], fn[20:22])
    return Time(dtisot, format="isot").jd




import os
import numpy as np
from astropy.io import fits
from collections import OrderedDict

__all__ = ["CCD", ]


class CCD(np.ndarray):
    """ CCD class, to cope with ccd images

    Note
    ----
    This class is a subclass of numpy.ndarray.

    References
    ----------
    https://docs.scipy.org/doc/numpy/user/basics.subclassing.html
    https://docs.scipy.org/doc/numpy/reference/arrays.classes.html

    Updates
    -------
    remove mean, median and std methods.

    """

    # ccd info
    gain = 1
    ron = 0
    unit = "adu"

    # file info, only when read from fits file
    fp = ""
    hdu = 0
    header = None
    meta = OrderedDict()

    # post-processing info
    _trim = None
    _rot90 = 0

    extr_attr_list = ["gain", "ron", "unit",            # ccd info
                      "fp", "hdu", "header", "meta",    # file info
                      "_trim", "_rot90"]                # post-proc info

    ###########################
    # print info
    ###########################
    def pprint(self):
        s = ("< twodspec.ccd.CCD instance {} x {} >\n"
             "gain  = {}\n"
             "ron   = {}\n"
             "unit  = {}\n"
             "trim  = {}\n"
             "rot90 = {}\n"
             "fp    = {}\n"
             "{}".format(*self.shape,
                         self.gain, self.ron, self.unit, self._trim,
                         self._rot90, self.fp,
                         self.__repr__()))
        print(s)
        return

    @property
    def info(self):
        """ print basic info

        Usage:
        ------
        CCD.info

        """
        s = ("< twodspec.ccd.CCD instance {} x {} >\n"
             "gain  = {}\n"
             "ron   = {}\n"
             "unit  = {}\n"
             "trim  = {}\n"
             "rot90 = {}\n"
             "fp    = {}\n"
             "hdu   = {}".format(*self.shape,
                                 self.gain, self.ron, self.unit, self._trim,
                                 self._rot90, self.fp, self.hdu))
        print(s)
        return

    ###########################
    # to inherit numpy.ndarray
    ###########################
    def __new__(cls, data, gain=1., ron=0., unit="adu",
                trim=None, rot90=0, header=None, meta=None,
                offset=0, strides=None, order="C", info=None):
        # Create the ndarray instance of our type, given the usual
        # ndarray input arguments.  This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to InfoArray.__array_finalize__

        data = np.array(data, dtype=np.float, order=order)

        # substantiate
        ccd = super(CCD, cls).__new__(cls, shape=data.shape,
                                      dtype=data.dtype, buffer=data,
                                      offset=offset, strides=strides,
                                      order=order)
        # set info
        ccd.gain = gain
        ccd.ron = ron
        ccd.unit = unit
        ccd._trim = trim
        ccd._rot90 = rot90
        ccd.header = header
        if header is None:
            ccd.meta = OrderedDict()
        else:
            ccd.meta = OrderedDict(header)

        return ccd

    # def __array_finalize__(self, obj):
    #     if obj is None:
    #         return

    # def __array_wrap__(self, out_arr, context=None):
    #     print('In __array_wrap__:')
    #     print('   self is %s' % repr(self))
    #     print('   arr is %s' % repr(out_arr))
    #     # then just call the parent
    #     return super(CCD, self).__array_wrap__(self, out_arr, context)

    ###########################
    # read from fits
    ###########################
    @staticmethod
    def reads(fps=["",], method="median", std=False, hdu=0, gain=1., ron=0., unit="adu", trim=None, rot90=0):
        """ read and combine multiple ccd frames """
        # read
        ccds = CCD([CCD.read(fp, hdu=hdu, gain=gain, ron=ron, unit=unit, trim=trim, rot90=rot90) for fp in fps])

        # combine
        if method == "median":
            ccd_comb = np.median(ccds, axis=0)
        elif method == "mean":
            ccd_comb = np.mean(ccds, axis=0)
        else:
            raise ValueError("@CCD.combine: bad method [{}]".format(method))
        # set info
        ccd_comb.gain = gain
        ccd_comb.ron = ron
        ccd_comb.unit = unit
        ccd_comb._trim = trim
        ccd_comb._rot90 = rot90

        if not std:
            return ccd_comb
        else:
            # evaluate std
            ccd_std = np.std(ccds, axis=0)
            # set info
            ccd_std.gain = gain
            ccd_std.ron = ron
            ccd_std.unit = unit
            ccd_std._trim = trim
            ccd_std._rot90 = rot90
            return ccd_comb, ccd_std

    @staticmethod
    def read(fp="", hdu=0, gain=1., ron=0., unit="adu", trim=None, rot90=0):
        if os.path.exists(fp):
            # read fits
            hdu_list = fits.open(fp)

            # get data
            data = hdu_list[hdu].data
            header = hdu_list[hdu].header
            meta = OrderedDict(header)

            # trim
            if trim is not None:
                l, r, t, b = trim
                data = data[l:r + 1, t:b + 1]

            # rotate
            data = np.rot90(data, rot90)

            # initialize CCD
            ccd = CCD(data, gain=gain, ron=ron, unit=unit,
                      trim=trim, rot90=rot90, header=header, meta=meta)
            ccd.fp = fp

            return ccd

        else:
            raise ValueError("@CCD.read(): file not found! [{}]".format(fp))
    
    def trim(self, trim=(0, 2048, 0, 2048)):
        """ trim CCD data

        Parameters
        ----------
        trim : [left, right, top, bottom]
            the trimed section
        """
        l, r, t, b = trim
        trimed_data = self[l:r, t:b]
        trimed_data.copy_info(self)
        return trimed_data

    def rot90(self, k):
        """ Rotate self by 90 degrees in the counter-clockwise direction. <-\

        Parameters
        ----------
        k : integer
            Number of times the array is rotated by 90 degrees.
        """
        rotated_data = np.rot90(self, k)
        rotated_data.copy_info(self)
        return rotated_data

    ###########################
    # copy option
    ###########################
    def copy_info(self, ccd1):
        """ copy info from an other CCD instance """
        for k in self.extr_attr_list:
            self.__setattr__(k, ccd1.__getattribute__(k))

    # def copy(self):
    #     return np.copy(self)

    ###########################
    # get config
    ###########################
    @property
    def config(self):
        return dict(hdu=self.hdu,
                    gain=self.gain,
                    ron=self.ron,
                    unit=self.unit,
                    trim=self._trim,
                    rot90=self._rot90)

    def write(self, fp, overwrite=True):
        phdu = fits.PrimaryHDU(data=self, header=fits.Header(self.meta))
        phdu.writeto(fp, overwrite=overwrite)
        return

    ###########################
    # arithmetic options
    ###########################
    # def subtract(self, ccd1):
    #     """ ccd2 = self - ccd1 """
    #     ccd2 = self/ccd1
    #     ccd2.copy_info(self)
    #     return ccd2
    #
    # def devide(self, ccd1):
    #     """ ccd2 = self / ccd1 """
    #     ccd2 = self / ccd1
    #     ccd2.copy_info(self)
    #     return ccd2

    # @staticmethod
    # def combine(ccds, method="median"):
    #     """ combine ccd frames """
    #     ccds = CCD(ccds)
    #     if method == "median":
    #         return np.median(ccds, axis=0)
    #     elif method == "mean":
    #         return np.mean(ccds, axis=0)
    #     else:
    #         raise ValueError("@CCD.combine: bad method [{}]".format(method))

    # def mean(self, axis=None):
    #     return CCD(np.mean(self, axis=axis))
    #
    # def median(self, axis=None):
    #     return CCD(np.median(self, axis=axis))
    #
    # def std(self, axis=None):
    #     return CCD(np.std(self, axis=axis))


def test2():
    fp = os.getenv("HOME") + "/PycharmProjects/songcn/twodspec/data/s2_2017-01-13T16-43-05.fits"
    # check image --> consistent with fits.getdata & ccdproc.CCDData.read
    print("---")
    data = fits.getdata(fp)
    print(data)

    print("---")
    from ccdproc import CCDData
    ccddata = CCDData.read(fp, unit="adu")
    print(ccddata)

    print("---")
    for i in range(4):
        ccd = CCD.read(fp, rot90=i, gain=5)
        print(ccd)
    print("---")


def test1():
    fp = os.getenv("HOME") + "/PycharmProjects/songcn/twodspec/data/s2_2017-01-13T16-43-05.fits"
    ccd = CCD.read(fp, rot90=0, gain=5)
    print(ccd)

    print(ccd.info)
    print(ccd.config)

    from matplotlib import pyplot as plt
    plt.figure()
    plt.imshow(ccd, origin="upper")


if __name__ == "__main__":
    test1()
    test2()



# #!/usr/bin/env python2
# # -*- coding: utf-8 -*-
# """
# Created on Mon Feb 13 12:35:08 2017
#
# @author: cham
# """
#
# #%%
# %pylab qt
# %matplotlib qt
# %load_ext autoreload
# %autoreload 2
#
#
# from song import master_
# from song import measure_xdrift
# from song import ccdproc_mod
# from song import utils
# import hrs
#
# from matplotlib import rcParams
# rcParams.update({'font.size':20})
# #%%
#
# reload(master_)
# reload(measure_xdrift)
#
# #%%
#
# # specify directory
# dp = '/hydrogen/song/star_spec/20170102/night/raw/'
#
# # scan fits files
# t = measure_xdrift.scan_files(dp)
#
# # measure cross-order drift
# t2, fig = measure_xdrift.check_xdrift(t)
# f90 = utils.scan_flux90(t, n_jobs=20, unit='adu')
# plot(f90[t['IMAGETYP']=='FLAT'])
# plot(f90[t['IMAGETYP']=='STAR'])
#
#
# """ BIAS """
# bias = master_.produce_master(t2, method='average', imagetp='BIAS', slc=slice(0, 10)).rot90(1)
#
# """ FLAT """
# flat = master_.produce_master(t2, method='average', imagetp='FLAT', slc=slice(0, 100)).rot90(1)
# flat_bias = ccdproc_mod.subtract_bias(flat, bias)
#
# """ THAR """
# thar = master_.produce_master(t2, method='median', imagetp='THAR', slc=slice(0, 1)).rot90(1)
# thar_bias = ccdproc_mod.subtract_bias(thar, bias)
#
#
#
# """ list all STAR images """
# master_.list_image(t2, imagetp='STAR', kwds=['OBJECT'])
#
#
#
# #%%
# """ VCUT """
# flat216 = ccdproc_mod.CCDData.read('/town/HRS/20161110/20161110006.fits', unit='adu')
# thar216 = ccdproc_mod.CCDData.read('/town/HRS/20161110/201611100039.fits', unit='adu')
# star216 = ccdproc_mod.CCDData.read('/town/HRS/20161110/201611100052.fits', unit='adu')
#
# fig = figure(figsize=(15, 10))
# ax = fig.add_subplot(211)
# ax.plot(flat_bias[:,1024])
# ax.plot(thar_bias[:,1024])
# ax.plot(star_bias[:,1024])
# ax.legend(['flat', 'thar', 'star'])
# ax.set_title('SONG')
#
# ax = fig.add_subplot(212)
# ax.plot(flat216[:,2048])
# ax.plot(thar216[:,2048])
# ax.plot(star216[:,2048])
# ax.legend(['flat', 'thar', 'star'])
# ax.set_title('2.16m HRS')
#
#
#
# #%%
#
# """ read noise """
# fps_bias_top10 = master_.list_image(t2, imagetp='BIAS', return_col='fps', kwds=['OBJECT']).data
#
# bias_top10 = []
# for fp in fps_bias_top10:
#     bias_top10.append(ccdproc_mod.CCDData.read(fp, unit='adu')).data
#
# bias_top10 = np.array(bias_top10)
# bias_top10_std = np.std(bias_top10, axis=0)
#
#
# imshow(bias_top10_std, interpolation='nearest')
#
#
#
# #%%
#
#
# # #################################### #
# #    find apertures
# # #################################### #
#
# #extract.list_image(t, 'FLAT')
#
# """ find apertures """
# find_aps_param_dict = dict(start_col=440, max_drift=8, max_apwidth=10,
#                            n_pix_goodap=100, n_adj=10, n_smooth=1, n_sep=3, c=5)
# ap_comb = hrs.combine_apertures([flat_bias], n_jobs=1, find_aps_param_dict=find_aps_param_dict)
# cheb_coefs, ap_uorder_interp = hrs.group_apertures(ap_comb, start_col=1024, order_dist=7)
#
# """ extract 1d FLAT & STAR & THAR """
# flat_bias_sl = hrs.substract_scattered_light(flat_bias, ap_uorder_interp, ap_width=10, shrink=.85)
# flat1d = hrs.extract_1dspec(flat_bias_sl, ap_uorder_interp, ap_width=8)[0]
#
# thar_bias_sl = hrs.substract_scattered_light(thar_bias, ap_uorder_interp, ap_width=10, shrink=.85)
# thar1d = hrs.extract_1dspec(thar_bias, ap_uorder_interp, ap_width=8)[0]
#
#
#
# # #################################### #
# #    determine wave_init
# # #################################### #
#
# # laod template thar
# thar_temp_path = '/home/cham/PycharmProjects/hrs/song/calibration/thar_template/thar_template.fits'
# wave_temp, thar_temp, order_temp = hrs.load_thar_temp(thar_temp_path)
#
# # fix thar
# thar1d_fixed = hrs.fix_thar_sat_neg(thar1d, arm=30, sat_count=60000*15)
# thar_temp_fixed = hrs.fix_thar_sat_neg(thar_temp, arm=30, sat_count=60000*15)
#
#
# """ initial estimation of wavelength """
# ytrim = (wave_temp.shape[0]*np.array([.2, .8])).astype(int)
# xtrim = (wave_temp.shape[1]*np.array([.2, .8])).astype(int)
# shift, corr2d = hrs.thar_corr2d(thar1d_fixed-np.percentile(thar1d_fixed, 50),
#                                 thar_temp-np.percentile(thar_temp, 50),
#                                 xtrim=xtrim, ytrim=ytrim,
#                                 y_shiftmax = 3, x_shiftmax=20)
#
# imshow(corr2d, interpolation='nearest')
# print ("@Cham: the shift is ", shift)
#
# wave_init = hrs.interpolate_wavelength_shift(wave_temp, shift, thar_temp, thar1d_fixed)
# order_init = hrs.interpolate_order_shift(order_temp, shift, thar1d_fixed)
#
#
# # #################################### #
# #    determine wave_init
# # #################################### #
#
# """ list all STAR images """
# master_.list_image(t2, imagetp='STAR', kwds=['OBJECT'])
#
#
# """ STAR """
# fig = figure()
#
# for i in range(63):
#     star = master_.produce_master(t2, method='median', imagetp='STAR', slc=slice(i, i + 1)).rot90(1)
#     star_bias = ccdproc_mod.subtract_bias(star, bias)
#
#     star_bias_sl = hrs.substract_scattered_light(star_bias, ap_uorder_interp, ap_width=10, shrink=.85)
#     star1d = hrs.extract_1dspec(star_bias_sl, ap_uorder_interp, ap_width=8)[0]
#
#     star1d_dblz = star1d/flat1d
#
#     #plot(wave_init.T, (flat1d/np.percentile(flat1d, 90, axis=1)[:, None]).T, 'b')
#     #plot(wave_init.T, (star1d/np.percentile(star1d, 90, axis=1)[:, None]).T, 'g')
#     plot(wave_init.T, (star1d_dblz/np.percentile(star1d_dblz, 90, axis=1)[:, None]).T*0.9+i, 'r')
#
#
#
# fig = figure()
# plot(wave_temp.T, thar_temp.T, 'r')
# plot(wave_init.T, thar1d.T, 'b')
#
#
# apflat = ccdproc_mod.CCDData.read('/home/cham/s2_2017-01-02T19-42-30_.fits', unit='adu')
# figure(), imshow(apflat, vmin=0.5, vmax=1.5)
#
#
# np.savetxt
#
#
#
#
#
#
# plot(wave_init.T, star1d_dblz.T, 'r')
#
# np.savetxt('/home/cham/Desktop/thar1d.dat', thar1d)
#
# thar_threshold = .90
#
# for i in range(thar1d.shape[0]):
#     plot(thar1d[i]/np.max(thar1d[i])+i)
#     ind_gt_thar_thre = np.where(thar1d[i]>np.percentile(thar1d[i], thar_threshold*100))[0]
#     plot(ind_gt_thar_thre, (thar1d[i]/np.max(thar1d[i]))[ind_gt_thar_thre]+i, 'r.')
#
#
# hist(thar1d[0], 1000)
# plot(thar1d[0])
#
# """  """
# thar_temp_path = '/home/cham/PycharmProjects/hrs/hrs/calibration/thar_template/thar_temp_w20160120022t.fits'
# wave_temp, thar_temp, order_temp = load_thar_temp(thar_temp_path)
#
#
#
#
# #%%
# for i in range(star1d_dblz.shape[0]):
#     plot(star1d_dblz[i, :]/np.median(star1d_dblz[i, :])+i*1)
#
#
# #%%
# imshow(log10(thar_bias), interpolation='nearest')
# plot(ap_comb.T, 'k-')
# plot(ap_uorder_interp.T, 'w-')
#



# -*- coding: utf-8 -*-
"""

Author
------
Bo Zhang

Email
-----
bozhang@nao.cas.cn

Created on
----------
- Sun Jan  8 14:00:00 2017

Modifications
-------------
-

Aims
----
- utils for scanning SONG files

"""

import datetime
import sys
import time

import numpy as np
from astropy.io import fits
from astropy.nddata import CCDData
from astropy.table import Table, Column
from ipyparallel import Client
from joblib import Parallel, delayed
from tqdm import trange


def unique_type(data):
    """ get the unique types of a list of data """
    tps = [type(_) for _ in data]

    tpunique = []
    while len(tps)> 0:
        tp = tps[0]
        tpunique.append(tp)
        tpcount = tps.count(tp)
        for i in range(tpcount):
            tps.remove(tp)
    return tpunique


# grab fits header
def grab_fits_header(fps, n_jobs=2, verbose=True, *args, **kwargs):
    """ grab the fits headers for multiple files: fps """
    hs = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(fits.getheader)(fp, *args, **kwargs) for fp in fps)

    # get all keys
    all_keys = list(hs[0].keys())
    for i in range(len(hs)):
        for k in hs[i].keys():
            if k not in all_keys:
                all_keys.append(k)

    # columnize data
    columns = [Column(fps, 'fps')]
    for k in all_keys:
        this_column_data = []
        for h in hs:
            if k in h.keys():
                this_column_data.append(h[k])
            else:
                this_column_data.append(None)
        columns.append(Column(this_column_data, k))

    # change object to str
    columns_updated = []
    for i in range(len(columns)):
        if columns[i].dtype is np.dtype(object):
            # if dtype is object
            uniquetp = unique_type(columns[i].data)
            uniquetp.remove(type(None))
            if len(uniquetp) == 1:
                finaltp = uniquetp[0]
                # print(finaltp)
                if finaltp is int:
                    # int, replace None with -9999
                    data = []
                    for v in columns[i].data:
                        if v is None:
                            data.append(-9999)
                        else:
                            data.append(v)
                    columns_updated.append(
                        Column(data, columns[i].name, dtype=finaltp))
                else:
                    # float, etc
                    columns_updated.append(
                        Column(columns[i].data, columns[i].name,
                               dtype=finaltp))
            else:
                # if type mixed, use string
                columns_updated.append(
                    Column(columns[i].data, columns[i].name,
                           dtype=str))
        else:
            # if not object, use it directly
            columns_updated.append(columns[i])

    return Table(columns_updated)


# scan fits header
def scan_fits_header(fps, verbose=True):
    """ scan fits headers, with tqdm process bar """
    cfn = sys._getframe().f_code.co_name

    if verbose:
        # use tgdm.trange
        for i in trange(len(fps), ncols=100, ascii=False,
                        desc="**" + cfn + "**", unit=" file"):
            fp = fps[i]

            # read fits
            hl = fits.open(fp)
            hdr_dict = dict(hl[0].header)

            # initiate table
            if i == 0:
                t = Table()
                for k, v in hdr_dict.items():
                    if type(v) is str:
                        t.add_column(
                            Column(data=np.array([]), dtype=object, name=k))
                    else:
                        t.add_column(
                            Column(data=np.array([]), dtype=type(v), name=k))
            # add row
            t.add_row(hdr_dict.values())
    else:
        for i in range(len(fps)):
            fp = fps[i]

            # read fits
            hl = fits.open(fp)
            hdr_dict = dict(hl[0].header)

            # initiate table
            if i == 0:
                t = Table()
                for k, v in hdr_dict.items():
                    if type(v) is str:
                        t.add_column(
                            Column(data=np.array([]), dtype=object, name=k))
                    else:
                        t.add_column(
                            Column(data=np.array([]), dtype=type(v), name=k))
            # add row
            t.add_row(hdr_dict.values())

    # add fps column
    t.add_column(Column(fps, "fps"), 0)

    # fix column dtype
    for i, colname in enumerate(t.colnames):
        if t[colname].dtype == np.dtype("O"):
            newcol = t[colname].astype("str")
            t.remove_column(colname)
            t.add_column(newcol, i)

    return t


def scan_files(fps, n_jobs=2, verbose=True):
    """ scan fits file headers under dirpath """
    fps.sort()
    t = grab_fits_header(fps, n_jobs=n_jobs, verbose=verbose)
    return t


def scan_flux90(t, ind_scan=None, ind_type='bool', pct=90,
                n_jobs=-1, verbose=10, **kwargs):
    """

    Parameters
    ----------
    t: table
        catalog of files
    ind_scan: bool array

    ind_type: string
        {'bool', 'int'}
    pct: float
        default is 90
    n_jobs: int
        number of processes launched
    verbose:
        verbose level
    kwargs:
        ccdproc.CCDData.read() keywords

    Returns
    -------
    f90: array
        the 90th percentile array

    """
    if ind_scan is None:
        ind_scan = np.ones((len(t),), dtype=bool)
        ind_type = 'bool'

    f90_ = Parallel(n_jobs=n_jobs, verbose=verbose)(
        delayed(scan_flux90_)(fp, pct, **kwargs) for fp in t['fps'][ind_scan])

    if ind_type == 'bool':
        f90 = np.array(f90_)*ind_scan
    else:
        f90 = np.zeros((len(t),), dtype=float)
        for i90_, i90 in enumerate(ind_scan):
            f90[i90] = f90_[i90_]

    return f90


def scan_flux90_(fp, pct, **kwargs):
    return np.nanpercentile(CCDData.read(fp, **kwargs), pct)


def keep_ipc_alive(profile="default", interval=60, maxiter=10000):
    dv = Client(profile=profile)[:]
    c = 0
    while True:
        c += 1
        print("@keep_ipc_alive: {} ".format(datetime.datetime.now()))
        dv.execute("import os")
        time.sleep(interval)
        if c > maxiter:
            break
    return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun May 21 12:28:52 2017

@author: Bo Zhang (bozhang@nao.cas.cn)

This module is to wrap some of the functions in STELLA (Ritter et al. 2014),
e.g., optimal extraction.

The scheme is to run the C executables, such as optextract through subprocess.Popen().

"""

import os
import subprocess
from tempfile import NamedTemporaryFile

import ccdproc
import numpy as np
from astropy.io import fits

from .aprec import AprecList

COMMAND_LIST = {"optextract", "extractsum", "makenormflat"}

# ######################### #
# simple wrappers of STELLA #
# ######################### #


def command_assert(cmd):
    """ assert a command in the list of wrapped commands """
    assert cmd in COMMAND_LIST


def print_stdout(bts):
    """ print stdout on screen """
    str_list = bts.decode("utf-8").split("\n")
    for str_ in str_list:
        print(str_)
    return


def command_help(cmd='optextract', shell=False):
    """
    
    Parameters
    ----------
    cmd: string
        command name
    shell: bool
        if True, shell function is enabled

    Returns
    -------
    
    Examples
    --------
    >>> command_help('extractsum')
    
    """
    assert cmd in COMMAND_LIST
    sp = subprocess.run(cmd, shell=shell, stdout=subprocess.PIPE)
    print_stdout(sp.stdout)
    return


def command_run(cmd, **kwargs):
    """ to run a bash command
    
    Parameters
    ----------
    cmd: 
        command and args
    kwargs: 
        keyword arguments passed to subprocess.run()

    Returns
    -------

    """
    s = subprocess.run(cmd, stdout=subprocess.PIPE, **kwargs)
    if s.returncode == 0:
        # success
        return 
    else:
        # fail
        print_stdout(s.stdout)
        raise (RuntimeError(
            "@Cham: something went wrong when executing command '%s'" % (
            " ".join(s.args))))


def value2string(val):
    """ to convert val to string """
    if val is None:
        return ""
    else:
        return str(val)

    
def args2command(*args):
    """ to convert positional arguments to string list """
    try:
        assert None not in args
        assert "" not in args        
    except:
        print("args:", args)
        raise(ValueError("None values not allowed in args!"))
    return [str(_).strip() for _ in args]


def kwargs2command(**kwargs):
    """ to convert keyword arguments to string list """
    cmd = []
    for k, v in kwargs.items():
        if v is not None:
            cmd.append("{}={}".format(str(k).strip(), str(v).strip()))
    return cmd


# ############################# #
#    wrapped STELLA commands    #
# ############################# #

def extractsum(*args, **kwargs):
    """
    USAGE: extractsum
    <char[] FitsFileName_In>
    <char[] DatabaseFileName_In> 
    <char[] FitsFileName_Out> 
    <double Gain> 
    [ERR_IN=char[]] 
    [ERR_OUT_EC=char[]] 
    [AREA=[int(xmin),int(xmax),int(ymin),int(ymax)]] 
    [APERTURES=<char[] ApertureFile_In]
    """
    if len(args) + len(kwargs) == 0:
        command_help("extractsum")
        return
    cmd = ["extractsum"]
    cmd.extend(args2command(*args))
    cmd.extend(kwargs2command(**kwargs))
    print("@stella.py: executing command ""{}""".format(cmd))
    command_run(cmd)
    return


def optextract(*args, **kwargs):
    """
    USAGE: optextract
    <char[] (@)FitsFileName_In>
    <char[] (@)DatabaseFileName_In>
    <char[] (@)FitsFileName_Out>
    <double Gain>
    <double ReadOutNoise>
    [TELLURIC=int[0 - none, 1 - Piskunov, 2 - LinFit]]
    [MAX_ITER_SF=int] 
    [MAX_ITER_SKY=int]
    [MAX_ITER_SIG=int]
    [SWATH_WIDTH=int] 
    [SF_SMOOTH=double] 
    [SP_SMOOTH=int] 
    [WING_SMOOTH_FACTOR=double] 
    [ERR_IN=char[](@)] 
    [ERR_OUT_2D=char[](@)] 
    [ERR_OUT_EC=char[](@)] 
    [ERRFIT_OUT_EC=char[](@)] 
    [SKY_OUT_EC=char[](@)]
    [SKYFIT_OUT_EC=char[](@)] 
    [SKY_OUT_2D=char[](@)] 
    [SKYFIT_OUT_2D=char[](@)]
    [SKY_ERR_OUT_EC=char[](@)] 
    [SKYFIT_ERR_OUT_EC=char[](@)]
    [PROFILE_OUT=char[](@)] 
    [IM_REC_OUT=char[](@)] 
    [REC_FIT_OUT=char[](@)] 
    [MASK_OUT=char[](@)] 
    [SPFIT_OUT_EC=char[](@)] 
    [EC_FROM_PROFILE_OUT=char[](@)]
    [AREA=[int(xmin),int(xmax),int(ymin),int(ymax)]]
    [XCOR_PROF=int] 
    [APERTURES=char[](@)]

    FitsFileName_In: image to extract
    DatabaseFileName_In: aperture-definition file to use for extraction
    FitsFileName_Out: output filename containing extracted spectra
    Gain: CCD gain
    ReadOutNoise: CCD readout noise
    TELLURIC: 0 - none, 1 - Piskunov, 2 - LinFit
    MAX_ITER_SF: maximum number of iterations calculating the slit function 
        (spatial profile)
    MAX_ITER_SKY: maximum number of iterations calculating the sky 
        (TELLURIC = 2 only)
    MAX_ITER_SIG: maximum number of iterations rejecting cosmic-ray hits
    SWATH_WIDTH: width of swath (bin) for which an individual profile shall be
        calculated
    SMOOTH_SF: Width of median SlitFunc smoothing
    SMOOTH_SP: Width of median Spectrum/Blaze smoothing
    WING_SMOOTH_FACTOR: Width of median SlitFunc-Wing smoothing
    ERR_IN: input image containing the uncertainties in the pixel values of 
        FitsFileName_In
    ERR_OUT_2D: output uncertainty image - same as ERR_IN, but with detected 
        cosmic-ray hits set to 10,000
    ERR_OUT_EC: output file containing the uncertainties in the extracted 
        spectra's pixel values from ExtractErrors
    ERRFIT_OUT_EC: output file containing the uncertainties in the extracted 
        spectra's pixel values from Fit
    SKY_OUT_EC: output sky-only spectra (TELLURIC > 0 only)
    SKYFIT_OUT_EC: output sky-only spectra (TELLURIC > 0 only)
    SKY_OUT_2D: reconstructed sky-only image
    SKYFIT_OUT_2D: reconstructed sky-only image
    SKY_ERR_OUT_EC: uncertainties in the calculated sky-only values 
        (only differs from SKYFIT_ERR_OUT_EC if TELLURIC==1
    SKYFIT_ERR_OUT_EC: uncertainties in the calculated sky-only values from Fit
    PROFILE_OUT: reconstructed image of the spatial profiles
    IM_REC_OUT: reconstructed input image from the profile-fitting/extraction
    SPFIT_OUT_EC: extracted spectra from linear fit of spatial profiles to 
        input spectra with 3-sigma rejection (ignoring mask), with sky if 
        TELLURIC>0, without sky if TELLURIC=0
    REC_FIT_OUT: reconstructed input image for SPFIT_OUT_EC
    MASK_OUT: output mask with detected cosmic-ray hits set to 0, good pixels 
        set to 1
    EC_FROM_PROFILE_OUT: extracted spectra from simply multiplying the input 
        image with the profile image as weight and summing up
    AREA: Area from which to extract spectra if center of aperture is in 
        specified area
    XCOR_PROF: How many cross-correlation runs from -1 pixel to +1 pixel 
        compared to XCenter?
    APERTURES: input filename containing a list of apertures to extract
    """
    if len(args) + len(kwargs) == 0:
        command_help("optextract")
        return
    cmd = ["optextract"]
    cmd.extend(args2command(*args))
    cmd.extend(kwargs2command(**kwargs))
    print("@stella.py: executing command ""{}""".format(cmd))
    command_run(cmd)
    return


def makenormflat(*args, **kwargs):
    """
    
    USAGE: makenormflat 
    <char[] FitsFileName_In> 
    <char[] DatabaseFileName_In> 
    <char[] NormalisedFlat_Out> 
    <char[] ProfileFits_Out> 
    <char[] BlazeFits_Out> 
    <double Gain> 
    <double ReadOutNoise> 
    <int SmoothSP> 
    <double MinSNR> 
    [<int SwathWidth>] 
    [ERR_IN=<char[] ErrFileName_In>]
    [SMOOTH_SF=<double smooth_slit_func>]
    [AREA=[<int xmin>,<int xmax>,<int ymin>,<int ymax>]]
    [MASK_OUT=<char[] Mask_Out>] 
    [RECONSTRUCTED_IMAGE_OUT=<char[] ReconstructedImage_Out>]
    Parameter 1: <char[] FitsFileName_In>: input flat fits file name
    Parameter 2: <char[] DatabaseFileName_In>: input Database file name
    Parameter 3: <char[] NormalisedFlat_Out>: output fits file name
    Parameter 4: <char[] ProfileFits_Out>: output profile image fits file name
    Parameter 5: <char[] BlazeFits_Out>: output blaze functions fits file name
    Parameter 6: <double Gain>: CCD gain
    Parameter 7: <double ReadOutNoise>: CCD read-out noise
    Parameter 8: <int SmoothSP>: smooth blaze function over SmoothSP pixels
    Parameter 9: <double MinSNR>: minimum S/N to use for each pixel of the 
        normalised flat. Pixels with lower S/N are set to unity
    Parameter 10: [<int SwathWidth>]: optional, length of swath to be used for 
        the determination of the spatial profile
    Parameter 11: [ERR_IN=<char[] ErrFileName_In>]: optional, name of fits file
        containing the uncertainty for the FitsFileName_In. If not set, the 
        errors will be calculated as the square root of FitsFileName_In
    Parameter 12: [SMOOTH_SF=<double smooth_slit_func>]: optional, smoothing 
        factor for the spatial profile (default = 1/10 with 10 pixels 
        oversampling)
    Parameter 13: [AREA=[<int xmin>,<int xmax>,<int ymin>,<int ymax>]]: 
        optional, area on the CCD to calculate and remove scattered light
    Parameter 14: [MASK_OUT=<char[] Mask_Out>]: optional, output fits file 
        containing the mask for the detected bad-pixels / cosmic-rays
    Parameter 15: [RECONSTRUCTED_IMAGE_OUT=<char[] ReconstructedImage_Out>]: 
        optional, output fits file containing the reconstructed flat
    example: makenormflat flat.fits database/apflat normalizedFlat.fits 
        flat_prof.fits flat_blaze.fits 3.8 10.5 5 100 300 SMOOTH_SF=0.5 
        MASK_OUT=flat_mask.fits RECONSTRUCTED_IMAGE_OUT=flat_rec.fits
    
    """
    if len(args) + len(kwargs) == 0:
        command_help("makenormflat")
        return
    cmd = ["makenormflat"]
    cmd.extend(args2command(*args))
    cmd.extend(kwargs2command(**kwargs))
    print("@stella.py: executing command ""{}""".format(cmd))
    command_run(cmd)
    return


# ##################################### #
# wrap STELLA functions at memory level #
# ##################################### #
PREFIX = "_twodspec_"


def write_image(im, filename=None, dir_work="", prefix="_twodspec_",
                suffix=".fits", delete=False):
    if filename is None:
        # tempfile in dir_work
        tf_im = NamedTemporaryFile(
            dir=dir_work, prefix=prefix, suffix=suffix, delete=delete)
        ccdproc.CCDData(im, unit='adu').write(tf_im.name, clobber=True)
        print("writing to {}".format(tf_im.name))
        tf_im.close()
        return tf_im.name
    else:
        # filename
        ccdproc.CCDData(im, unit='adu').write(filename, clobber=True)
        print("writing to {}".format(filename))
        return filename


def temp_image(dir_work=None):
    tf_im = NamedTemporaryFile(
        dir=dir_work, prefix=PREFIX, suffix=".fits", delete=False)
    return tf_im.name


def extractsum_prime(dir_work, im_in, database, err_in=None, err_out=None,
                     gain=1., **kwargs):
    """
    USAGE: extractsum
    <char[] FitsFileName_In>
    <char[] DatabaseFileName_In>
    <char[] FitsFileName_Out>
    <double Gain>
    [ERR_IN=char[]]
    [ERR_OUT_EC=char[]]
    [AREA=[int(xmin),int(xmax),int(ymin),int(ymax)]]
    [APERTURES=<char[] ApertureFile_In]
    """
    if dir_work[-1] != "/":
        dir_work += "/"

    if err_in is None:
        err_in = np.sqrt(np.abs(im_in))

    tf_im_in = write_image(im_in, filename=None, dir_work=dir_work)
    tf_im_out = temp_image(dir_work=dir_work)
    tf_err_in = write_image(err_in, filename=None, dir_work=dir_work)
    tf_err_out = temp_image(dir_work=dir_work)

    extractsum(tf_im_in, database, tf_im_out, gain,
               ERR_IN=tf_err_in, ERR_OUT_EC=tf_err_out, **kwargs)

    sp = ccdproc.CCDData.read(tf_im_out, unit="adu").data
    sp_err = ccdproc.CCDData.read(tf_err_out, unit="adu").data
    os.remove(tf_im_in)
    os.remove(tf_im_out)
    os.remove(tf_err_in)
    os.remove(tf_err_out)

    return sp, sp_err


def optextract_prime(dir_work, im_in, database, err_in=None, im_rec=None,
                     gain=1., ron=10., **kwargs):

    if dir_work[-1] != "/":
        dir_work += "/"

    if err_in is None:
        err_in = np.sqrt(np.abs(im_in))

    tf_im_in = write_image(im_in, filename=None, dir_work=dir_work)
    tf_im_out = temp_image(dir_work=dir_work)
    tf_err_in = write_image(err_in, filename=None, dir_work=dir_work)
    tf_err_out = temp_image(dir_work=dir_work)

    try:
        # try to extract all apertures
        if im_rec:
            tf_im_rec = temp_image(dir_work=dir_work)
            optextract(tf_im_in, database, tf_im_out, gain, ron,
                       ERR_IN=tf_err_in, ERR_OUT_EC=tf_err_out,
                       IM_REC_OUT=tf_im_rec, **kwargs)
            sp = ccdproc.CCDData.read(tf_im_out, unit="adu").data
            sp_err = ccdproc.CCDData.read(tf_err_out, unit="adu").data
            im_rec = ccdproc.CCDData.read(tf_im_rec, unit="adu").data
            os.remove(tf_im_in)
            os.remove(tf_im_out)
            os.remove(tf_err_in)
            os.remove(tf_err_out)
            os.remove(tf_im_rec)
            return sp, sp_err, im_rec
        else:
            optextract(tf_im_in, database, tf_im_out, gain, ron,
                       ERR_IN=tf_err_in, ERR_OUT_EC=tf_err_out, **kwargs)
            sp = ccdproc.CCDData.read(tf_im_out, unit="adu").data
            sp_err = ccdproc.CCDData.read(tf_err_out, unit="adu").data
            os.remove(tf_im_in)
            os.remove(tf_im_out)
            os.remove(tf_err_in)
            os.remove(tf_err_out)
            return sp, sp_err

    except RuntimeError as re_:
        # extract apertures one by one

        # split AprecList
        al = AprecList.read(database)
        al_fps = al.split(database+"_{:04d}")

        # extract each aperture
        sp = []
        sp_err = []
        for iap in range(al.nap):
            try:
                optextract(tf_im_in, al_fps[iap], tf_im_out, gain, ron,
                           ERR_IN=tf_err_in, ERR_OUT_EC=tf_err_out, **kwargs)
                sp.append(fits.getdata(tf_im_out).flatten())
                sp_err.append(fits.getdata(tf_err_out).flatten())

            except RuntimeError:
                print(tf_im_out)
                print(sp)
                print(sp_err)
                sp.append(None)
                sp_err.append(None)

        # replace None values
        isnotnone = [_ is not None for _ in sp]
        print("sp:", sp)
        if np.any(isnotnone):
            # default array
            defarr = sp[np.where(isnotnone)[0][0]] * np.nan
            for i in range(len(sp)):
                if sp[i] is None:
                    sp[i] = defarr
                    sp_err[i] = defarr
        else:
            print()
            raise RuntimeError("@STELLA: all apertures are bad!")

        return np.array(sp), np.array(sp_err)


def makenormflat_prime(dir_work, im_in, database, err_in=None, gain=1.,
                       ron=10., smoothsp=5, minsnr=15, bg=None, maxdev=None,
                       **kwargs):

    if dir_work[-1] != "/":
        dir_work += "/"

    if err_in is None:
        err_in = np.sqrt(np.abs(im_in))

    tf_im_in = write_image(im_in, filename=None, dir_work=dir_work)
    # tf_im_out = temp_image(dir_work=dir_work)
    tf_err_in = write_image(err_in, filename=None, dir_work=dir_work)
    tf_im_sst = temp_image(dir_work=dir_work)
    # tf_im_rec = temp_image(dir_work=dir_work)
    tf_im_prf = temp_image(dir_work=dir_work)
    tf_im_blz = temp_image(dir_work=dir_work)

    makenormflat(tf_im_in, database, tf_im_sst, tf_im_prf, tf_im_blz, gain,
                 ron, smoothsp, minsnr, ERR_IN=tf_err_in, **kwargs)
                 # RECONSTRUCTED_IMAGE_OUT=tf_im_rec,
    blaze = ccdproc.CCDData.read(tf_im_blz, unit="adu").data
    # im_rec_ = ccdproc.CCDData.read(tf_im_rec, unit="adu").data
    sensitivity = ccdproc.CCDData.read(tf_im_sst, unit="adu").data

    os.remove(tf_im_in)
    os.remove(tf_err_in)
    os.remove(tf_im_sst)
    # os.remove(tf_im_rec)
    os.remove(tf_im_prf)
    os.remove(tf_im_blz)

    if bg is None:
        # no background is added back
        if maxdev is not None:
            sensitivity = np.where(
                np.abs(sensitivity-1) >= maxdev, 1.0, sensitivity)
        return blaze, sensitivity
    else:
        # background is added back to calculate sensitivity
        sensitivity = np.where(
            sensitivity == 1.0, 1.0, im_in / ((im_in - bg) / sensitivity + bg))
        if maxdev is not None:
            sensitivity = np.where(
                np.abs(sensitivity-1) >= maxdev, 1.0, sensitivity)
        return blaze, sensitivity


# #%%
#
# command_run(['optextract','tmp_s2_2017-01-13T12-17-47.fits','masteraper_20170113_slit8.txt','tmp_s2_2017-01-13T12-17-47_ext.fits',' 1.0',' 10.0','ERR_IN=tmp_s2_2017-01-13T12-17-47.fits','SMOOTH_SF=0.1'])
#
#
# s = subprocess.run(["cat", ], stdout=subprocess.PIPE, input=b"abc")
#
# subprocess.run(["ls", "-l", "/dev/null"], stdout=subprocess.PIPE)
#
# cmd = [" tmp_s2_2017-01-13T12-17-47.fits", "masteraper_20170113_slit8.txt", "tmp_s2_2017-01-13T12-17-47_ext.fits", "1.0", "10.0", "ERR_IN=tmp_s2_2017-01-13T12-17-47.fits", "SMOOTH_SF=0.1"]
# cmd = ["optextract tmp_s2_2017-01-13T12-17-47.fits masteraper_20170113_slit8.txt tmp_s2_2017-01-13T12-17-47_ext.fits 1.0 10.0 ERR_IN=tmp_s2_2017-01-13T12-17-47.fits SMOOTH_SF=0.1"]
# cmd = ["optextract",]
#
#
# optextract(*cmd)
#
#
# s = subprocess.run(cmd, stdout=subprocess.PIPE)
# s = command_run(cmd)
# print_stdout(s.stdout)
#
# s = subprocess.run("optextract", shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
# s.stdout


# -*- coding: utf-8 -*-
"""
Created on Tue May 30 13:27:58 2017

@author: cham

@SONG: RMS =  0.00270124312246
delta_rv = 299792458/5500*0.00270124312246 = 147.23860278870518 m/s

LAMOST: R~1800,  299792.458/6000*3A = 150km/s WCALIB: 10km/s delta_rv = 5km/s
MMT: R~2500,  299792.458/R = 119.9km/s  rms=0.07A, delta_rv = 5km/s
SONG: 1800,  299792.458/6000*3A = 150km/s delta_rv = 5km/s
"""

import numpy as np
from astropy import table
from matplotlib import pyplot as plt
from scipy.interpolate import interp1d
from scipy.optimize import curve_fit
from scipy.optimize import minimize

from twodspec.polynomial import Poly2DFitter


def corr_thar(wave_temp, thar_temp, thar_obs, maxshift=100):
    """
    Calculate the shift between thar_obs and thar_temp using correlation.
    Returns the shifted interpolated wavelength array

    Parameters
    ----------
    wave_temp : ndarray
        wavelength template.
    thar_temp : ndarray
        thar template.
    thar_obs : ndarray
        thar observation.
    maxshift : int, optional
        max shift. The default is 100.

    Returns
    -------
    wave_corr : ndarray
        shifted interpolated wavelength array.

    """
    # assert number of orders are the same
    assert thar_obs.shape == thar_temp.shape
    nrow, ncol = thar_temp.shape
    icol0 = np.int(0.25 * ncol)
    icol1 = np.int(0.75 * ncol)
    assert icol0 > maxshift

    nshift_grid = np.arange(-maxshift, maxshift + 1)
    nshift_dotmax = np.zeros(nrow)
    for irow in np.arange(nrow):
        ind_dotmax = np.argmax(
            [np.dot(thar_temp[irow][icol0 + ishift:icol1 + ishift], thar_obs[irow][icol0:icol1]) for ishift in
             nshift_grid])
        nshift_dotmax[irow] = nshift_grid[ind_dotmax]
    bulkshift = np.median(nshift_dotmax)
    npm1 = np.sum(np.abs(nshift_dotmax - bulkshift) <= 1)
    npm2 = np.sum(np.abs(nshift_dotmax - bulkshift) <= 2)
    assert npm2 > 0.5 * nrow
    print("@corr_thar: bulkshift={} (±1 {}/{}) (±2 {}/{})".format(bulkshift, npm1, nrow, npm1, nrow))

    xcoord = np.arange(ncol)
    wave_corr = interp1d(xcoord - bulkshift, wave_temp, kind="linear", fill_value="extrapolate")(xcoord)
    return wave_corr


def ccfmax_gauss(x, b=0, c=1):
    """
    Generate a Gaussian array, centering at b, with width of c.

    Parameters
    ----------
    x : ndarray
        x array.
    b : float, optional
        The center of Gaussian. The default is 0.
    c : float, optional
        The width of Gaussian. The default is 1.

    Returns
    -------
    y
        The Gaussian array.

    """
    y = np.exp(-0.5 * ((x - b) / c) ** 2.)
    return y / np.sum(y)


def ccfmax_cost(x0, x, y, width=2):
    """
    The cost function for ccfmax

    Parameters
    ----------
    x0 : float
        initial guess of max position.
    x : array
        x.
    y : array
        y.
    width : float, optional
        the width of Gaussian. The default is 2.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    y_model = ccfmax_gauss(x, x0, width)
    return - np.dot(y, y_model)


def ccfmax(x0, x, y, width=2, method="Nelder-Mead"):
    """
    The ccf max method

    Parameters
    ----------
    x0 : float
        initial guess of max position.
    x : array
        x.
    y : array
        y.
    width : float, optional
        the width of Gaussian. The default is 2.
    method : str, optional
        minimization method. The default is "Nelder-Mead".

    Returns
    -------
    pccf : TYPE
        DESCRIPTION.

    """
    pccf = minimize(ccfmax_cost, x0=x0, args=(x, y, width), method=method)
    return pccf


def gauss(x, a, b, c):
    """
    generate a Gaussian function

    Parameters
    ----------
    x : array like
        x coordinates.
    a : float
        amplitude (centeral height).
    b : float
        center.
    c : float
        sigma.

    Returns
    -------
    y
        y = Gaussian(x | a, b, c).

    """
    c = np.abs(c)
    return a / np.sqrt(2. * np.pi) / c * np.exp(-0.5 * ((x - b) / c) ** 2.)


def find_lines(wave_init, thar_obs, thar_line_list, npix_chunk=20, ccf_kernel_width=2):
    """
    Find emission lines in ThAr spectrum.

    Parameters
    ----------
    wave_init : array
        initial wavelength solution.
    thar_obs : array
        observed ThAr spectrum.
    thar_line_list : array
        the ThAr line list (a list of wavelengths).
    npix_chunk : int, optional
        the chunk length (half). The default is 20.

    Returns
    -------
    tlines : astropy.table.Table
        A table of identified lines.

    """
    # get a chunk [5*5] ±5sigma
    # shift: 1-2 pixel
    # typical oversampling: 1/3R --> 3pixel=FWHM
    # LAMOST MRS overestimate: 0.7 / 0.1 --> 7pixel=FWHM
    norder, npix = wave_init.shape
    xcoord = np.arange(npix)

    tlines = []
    # for each order
    for iorder in range(norder):
        # this order
        this_wave_init = wave_init[iorder]
        this_wave_min = np.min(this_wave_init)
        this_wave_max = np.max(this_wave_init)
        this_thar_obs = thar_obs[iorder]
        this_line_list = thar_line_list[np.logical_and(thar_line_list > this_wave_min, thar_line_list < this_wave_max)]

        # for each line
        for this_line in this_line_list:
            # init x position
            this_line_x_init = np.interp(this_line, this_wave_init, xcoord)
            this_line_x_init_int = np.int(this_line_x_init)  # np.argmin(np.abs((this_wave_init-this_line)))

            # get a chunk
            if npix_chunk < this_line_x_init_int < npix - npix_chunk:
                this_line_slc = slice(this_line_x_init_int - npix_chunk, this_line_x_init_int + npix_chunk)
                this_line_xcoord = xcoord[this_line_slc]
                this_line_thar = this_thar_obs[this_line_slc]
                this_line_base = np.percentile(this_line_thar, q=20)  # 25th percentile as baseline
                # if this_line_base < 0:
                #     continue

                # 1. Gaussian fit
                try:
                    popt, pcov = curve_fit(gauss, this_line_xcoord, this_line_thar - this_line_base,
                                           p0=[100, this_line_x_init, 1.5], )
                    # bounds=(np.array([0,-np.inf,1]), np.array([np.inf,np.inf,np.inf])))
                    this_line_a_gf = popt[0]
                    this_line_c_gf = popt[2]
                    this_line_x_gf = popt[1]
                    this_line_wave_init_gf = np.interp(popt[1], xcoord, this_wave_init)
                except:
                    this_line_a_gf = np.nan
                    this_line_c_gf = np.nan
                    this_line_x_gf = np.nan
                    this_line_wave_init_gf = np.nan
                # 2. CCF method
                try:
                    pccf = ccfmax(this_line_x_init, this_line_xcoord, this_line_thar, width=ccf_kernel_width, method="Nelder-Mead")
                    this_line_x_ccf = np.float(pccf.x)
                    this_line_wave_init_ccf = np.interp(this_line_x_ccf, xcoord, this_wave_init)
                    this_line_peakflux = np.interp(this_line_x_ccf, this_line_xcoord, this_line_thar)
                except:
                    this_line_x_ccf = np.nan
                    this_line_wave_init_ccf = np.nan
                    this_line_peakflux = np.nan

                # gather results
                this_result = dict(
                    order=iorder,
                    line=this_line,
                    line_x_init=this_line_x_init,
                    # gf
                    line_x_gf=this_line_x_gf,
                    line_a_gf=this_line_a_gf,
                    line_c_gf=this_line_c_gf,
                    line_wave_init_gf=this_line_wave_init_gf,
                    # ccf
                    line_x_ccf=this_line_x_ccf,
                    line_wave_init_ccf=this_line_wave_init_ccf,
                    line_base=this_line_base,
                    # peakflux
                    line_peakflux=this_line_peakflux
                )
                # np.array([iorder, this_line_xcenter, this_line, line_wave_gf, line_wave_ccf,
                #           this_line_base, popt[0]/np.sqrt(2.*np.pi)/popt[2],
                #           *popt, np.float(pccf.x)]))
                tlines.append(this_result)
    tlines = table.Table(tlines)
    print("@find_lines: {}/{} lines using GF / CCF!".format(
        np.sum(np.isfinite(tlines["line_x_gf"])),
        np.sum(np.isfinite(tlines["line_x_ccf"]))
    ))

    return tlines


def grating_equation(x, y, z, deg=(4, 10), nsigma=3, min_select=None, verbose=True):
    """
    Fit a grating equation (2D polynomial function) to data

    Parameters
    ----------
    x : array
        x coordinates of emission lines
    y : array
        order number.
    z : array
        The true wavelengths of lines.
    deg : tuple, optional
        The degree of the 2D polynomial. The default is (4, 10).
    nsigma : float, optional
        The data outside of the nsigma*sigma radius is rejected iteratively. The default is 3.
    min_select : int or None, optional
        The minimal number of selected lines. The default is None.
    verbose :
        if True, print info

    Returns
    -------
    pf1, pf2, indselect

    """
    indselect = np.ones_like(x, dtype=bool)
    iiter = 0
    # pf1
    while True:
        pf1 = Poly2DFitter(x[indselect], y[indselect], z[indselect], deg=deg, pw=1, robust=False)
        z_pred = pf1.predict(x, y)
        z_res = z_pred - z
        sigma = np.std(z_res[indselect])
        indreject = np.abs(z_res[indselect]) > nsigma * sigma
        n_reject = np.sum(indreject)
        if n_reject == 0:
            # no lines to kick
            break
        elif isinstance(min_select, int) and min_select >= 0 and np.sum(indselect) <= min_select:
            # selected lines reach the threshold
            break
        else:
            # continue to reject lines
            indselect &= np.abs(z_res) < nsigma * sigma
            iiter += 1
        if verbose:
            print("@grating_equation: iter-{} \t{} lines kicked, {} lines left, rms={:.5f} A".format(
                iiter, n_reject, np.sum(indselect), sigma))
    pf1.rms = sigma

    # pf2
    pf2 = Poly2DFitter(x[indselect], y[indselect], z[indselect], deg=deg, pw=2, robust=False)
    pf2.rms = np.std(pf2.predict(x[indselect], y[indselect]) - z[indselect])
    if verbose:
        print("@grating_equation: {} iterations, rms = {:.5f}/{:.5f} A".format(iiter, pf1.rms, pf2.rms))
    return pf1, pf2, indselect


def test_corr_thar():
    import joblib
    wave_temp = joblib.load("/figs/wave_temp.dump")
    thar_temp = joblib.load("/figs/thar_temp.dump")
    thar_obs = joblib.load("/figs/thar_obs.dump")
    # load thar line list
    thar_line_list = joblib.load("/figs/thar_line_list.dump")
    # correlation for initial wavelength guess 
    wave_init = corr_thar(wave_temp, thar_temp, thar_obs, maxshift=100)
    # find thar lines
    tlines = find_lines(wave_init, thar_obs, thar_line_list, npix_chunk=20)

    # # initialize
    x = tlines["line_x_ccf"]
    y = tlines["order"]
    z = tlines["line"]

    # indfnt = np.isfinite(x)
    # x = x[indfnt]
    # y = y[indfnt]
    # z = z[indfnt]

    # fit grating equation
    pf1, pf2, indselect = grating_equation(x, y, z, deg=(4, 10), nsigma=3)
    tlines.add_column(table.Column(indselect, "indselect"))
    # predict wavelength solution
    nx, norder = thar_obs.shape
    mx, morder = np.meshgrid(np.arange(norder), np.arange(nx))
    wave_solu = pf2.predict(mx, morder)

    plt.figure()
    plt.plot(wave_temp.T, thar_temp.T, "b")
    plt.plot(wave_init.T, thar_obs.T, "r")
    plt.plot(wave_solu.T, thar_obs.T, "cyan")
    plt.vlines(thar_line_list, 0, 1e6)
    return

# %%
# deprecated!
# def calibrate(thar1d_simple, thar_solution_temp, thar_list,
#               poly_order=(5, 10), slit=5):
#     wave_temp, thar_temp, order_temp = thar_solution_temp
#     thar_temp += 400
#     thar_temp[thar_temp<0] = 0

#     wave_temp = np.flipud(wave_temp)
#     thar_temp = np.flipud(thar_temp)
#     # # 1.initial solution
#     # print("@SONG: [ThAr] 2D cross-correlation ...")
#     # shift, corr2d = calib.thar1d_corr2d(
#     #     thar1d_simple, thar_temp, y_shiftmax=3, x_shiftmax=40)
#     # figure();imshow(corr2d)

#     # print("@SONG: [ThAr] the shift is ", shift)
#     # wave_init = calib.interpolate_wavelength(
#     #     wave_temp, shift, thar_temp, thar1d_simple)
#     # order_init = calib.interpolate_order(
#     #     order_temp, shift, thar1d_simple) + 80

#     print("@SONG: [ThAr] refine wavelength ...""")
#     # 2.fit Gaussians to Thar lines
#     lc_coord, lc_order, lc_thar, popt, pcov = calib.refine_thar_positions(
#         wave_init, order_init, thar1d_simple, thar_list,
#         fit_width=.3, lc_tol=.1, k=3, n_jobs=-1, verbose=10)

#     # select using center deviation & line SNR
#     ind_good0 = np.logical_and(
#         np.abs(popt[:, 2] - lc_thar) < 5,  # (popt[:,3]*3),
#         (popt[:, 1] / np.sqrt(2. * np.pi) / popt[:, 3] / np.abs(
#             popt[:, 0])) > .1)
#     print(np.sum(ind_good0))

#     # 3.rejections of outliers
#     ind_good1 = calib.clean_thar_polyfit1d_reject(
#         lc_coord, lc_order, lc_thar, popt, ind_good0=ind_good0, deg=1, w=None,
#         epsilon=0.002, n_reserve=8)
#     print(np.sum(ind_good1))

#     # 4. fit final solution
#     print("@SONG: [ThAr] fit wavelength solution ")
#     x_mini_lsq, ind_good_thar, scaler_coord, scaler_order, scaler_ml = calib.fit_grating_equation(
#         lc_coord, lc_order, lc_thar, popt, pcov, ind_good_thar0=ind_good1,
#         poly_order=poly_order, max_dev_threshold=.003, n_iter=1, lar=False,
#         nl_eachorder=5)

#     # construct grids for coordinates & order
#     grid_coord, grid_order = np.meshgrid(np.arange(thar1d_simple.shape[1]),
#                                          np.arange(
#                                              thar1d_simple.shape[0]) + 80)

#     # 4'.fit grating function
#     sgrid_fitted_wave = calib.grating_equation_predict(
#         grid_coord, grid_order, x_mini_lsq, poly_order,
#         scaler_coord, scaler_order, scaler_ml)

#     # 5.get the fitted wavelength
#     lc_thar_fitted = calib.grating_equation_predict(
#         lc_coord, lc_order, x_mini_lsq, poly_order,
#         scaler_coord, scaler_order, scaler_ml)

#     results = [sgrid_fitted_wave, lc_thar_fitted]

#     # 6.figures for diagnostics
#     bins = np.arange(4500, 7500, 500)
#     bins_med, _, _ = binned_statistic(lc_thar[ind_good_thar],
#                                       lc_thar_fitted[ind_good_thar] - lc_thar[
#                                           ind_good_thar], statistic=np.median,
#                                       bins=bins)
#     bins_rms, _, _ = binned_statistic(lc_thar[ind_good_thar],
#                                       lc_thar_fitted[ind_good_thar] - lc_thar[
#                                           ind_good_thar], statistic=nanrms,
#                                       bins=bins)

#     """ [Figure]: calibration diagnostics """
#     fig = plt.figure(figsize=(12, 8))
#     fig.add_subplot(111)
#     plt.plot(lc_thar, lc_thar_fitted - lc_thar, '.')
#     plt.plot(lc_thar[ind_good_thar],
#              lc_thar_fitted[ind_good_thar] - lc_thar[ind_good_thar], 'r.')
#     plt.errorbar(bins[:-1] + np.diff(bins) * .5, bins_med, bins_rms, color='k',
#                  ecolor='k')
#     plt.xlim(4300, 7200)
#     plt.ylim(-.008, .008)
#     plt.xlabel("Wavelength (A)")
#     plt.ylabel("$\lambda(solution)-\lambda(true)$")
#     plt.title("RMS = {:05f} A for SLIT {:d} [{} lines]".format(
#         rms(lc_thar_fitted[ind_good_thar] - lc_thar[ind_good_thar]), slit,
#         len(ind_good_thar)))
#     plt.legend(["deviation of all lines", "deviation of used lines",
#                 "mean RMS in bins"])
#     fig.tight_layout()
#     # fig.savefig(dir_work+"thar{}_{:s}".format(slit, thar_fn.replace(".fits", "_diagnostics.svg")))
#     # plt.close(fig)
#     results.append(fig)

#     fig = plt.figure(figsize=(24, 8))
#     plt.imshow(np.log10(thar1d_simple), aspect='auto', cmap=cm.viridis,
#                vmin=np.nanpercentile(np.log10(thar1d_simple), 5),
#                vmax=np.nanpercentile(np.log10(thar1d_simple), 95))
#     plt.plot(lc_coord, lc_order - 80, ls='', marker='s', mfc='None', mec='b')
#     plt.plot(lc_coord[ind_good_thar], lc_order[ind_good_thar] - 80, ls='',
#              marker='s', mfc='None', mec='r')
#     plt.xlabel("CCD Coordinate")
#     plt.ylabel("Order")
#     plt.colorbar()
#     fig.tight_layout()
#     # fig.savefig(dir_work+"thar{}_{:s}".format(slit, thar_fn.replace(".fits", "_used_lines.svg")))
#     # plt.close(fig)
#     results.append(fig)

#     return results



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on %(date)s

@author: %(username)s
"""
#%%
import numpy as np
from astropy.io import fits
from laspec.normalization import normalize_spectrum_iter
from scipy import signal

from twodspec import extract
from twodspec.aperture import Aperture
from matplotlib.pyplot import figure, plot, imshow

#%%
def gaussian_kernel(sigma=2., n_sigma=5):
    """ generate Gaussian kernel

    Parameters
    ----------
    sigma : float, optional
        The sigma of the Gaussian. The default is 2..
    n_sigma : flaot, optional
        The width of the Gaussian in terms of sigma. The default is 5.

    Returns
    -------
    y : ndarray
        The Gaussian kernel array.

    """
    npix = np.int(n_sigma*sigma)+1
    x = np.arange(-npix, npix+1, 1)
    y = np.exp(-0.5*(x/sigma)**2)
    y /= np.sum(y)
    return y


def box_kernel(fullwidth=10):
    npix = np.int(fullwidth)
    kernel = np.ones(npix)
    kernel /= np.sum(kernel)
    return kernel

#%%
def trace_naive_max(flat, sigma=7., maxdev=7.):
    """ trace aperture using naive method -- along local max

    Parameters
    ----------
    flat : ndarray
        FLAT.
    sigma : float, optional
        The sigma of Gaussian kernel. The default is 7..
    maxdev : float, optional
        The max deviation of local max. The default is 7.

    Returns
    -------
    ap_center : ndarray (n_ap, n_pix)
        The array for center of aperture.

    """
    flat = np.asarray(flat, float)
    n_row, n_col = flat.shape
    i_row_center = np.int(n_row/2)
    
    # smooth flat along spacial axis
    flat_smooth = np.zeros_like(flat)
    gkernel = gaussian_kernel(sigma)
    for i_row in range(n_row):
        flat_smooth[i_row] = signal.fftconvolve(flat[i_row], gkernel, mode="same")
    
    # find local max
    flat_smooth_diff = np.diff(flat_smooth, axis=1)
    flat_localmax = np.zeros_like(flat, bool)
    flat_localmax[:,1:-1] = (flat_smooth_diff[:,1:]<0)&(flat_smooth_diff[:,:-1]>0)
    
    # start from center row
    ind_col_localmax, = np.where(flat_localmax[i_row_center])
    n_ap_try = len(ind_col_localmax)
    ap_center_try = np.ones((n_ap_try, n_row)) *np.nan
    ap_center_try[:, i_row_center] = ind_col_localmax
    
    # find local max col each row 
    each_row_localmax = [np.where(flat_localmax[i_row])[0] for i_row in range(n_row)]
    
    # loop for trials
    for i_ap in range(n_ap_try):
        for i_row in range(i_row_center-1, 0-1, -1):
            this_d = np.abs(each_row_localmax[i_row]-ap_center_try[i_ap, i_row+1])
            if np.min(this_d) <= sigma:
                ap_center_try[i_ap, i_row] = each_row_localmax[i_row][np.argmin(this_d)]
            else:
                break
        for i_row in range(i_row_center+1, n_row, 1):
            this_d = np.abs(each_row_localmax[i_row]-ap_center_try[i_ap, i_row-1])
            if np.min(this_d) <= sigma:
                ap_center_try[i_ap, i_row] = each_row_localmax[i_row][np.argmin(this_d)]
            else:
                break
    ind_good_ap = np.sum(np.isfinite(ap_center_try), axis=1) == n_row
    n_ap = np.sum(ind_good_ap)
    ap_center = ap_center_try[ind_good_ap]
    
    return ap_center

flat = fits.getdata("/Users/cham/projects/song/star_spec/20191105/night/ext/masterflat_20191105_slit5.fits")
ap_center = trace_naive_max(flat, sigma=7, maxdev=10)
plot(ap_center_try.T, np.arange(2048))
imshow(np.log10(flat))
print(ap_center.shape)
#%%
def trace_naive_min(flat, sigma=7., maxdev=5):
    flat = np.asarray(flat, float)
    n_row, n_col = flat.shape
    i_row_center = np.int(n_row/2)
    
    # smooth flat
    flat_smooth = np.zeros_like(flat)
    gkernel = gaussian_kernel(sigma)
    for i_row in range(n_row):
        flat_smooth[i_row] = signal.fftconvolve(flat[i_row], gkernel, mode="same")
    
    # find local max
    flat_smooth_diff = np.diff(flat_smooth, axis=1)
    flat_localmax = np.zeros_like(flat, bool)
    flat_localmax[:,1:-1] = (flat_smooth_diff[:,1:]>0)&(flat_smooth_diff[:,:-1]<0)
    
    # start from center row
    #maxdev = np.int(sigma)
    ind_col_localmax, = np.where(flat_localmax[i_row_center])
    n_ap_try = len(ind_col_localmax)
    ap_center_try = np.ones((n_ap_try, n_row)) *np.nan
    ap_center_try[:, i_row_center] = ind_col_localmax
    
    # each row local max col
    each_row_localmax = [np.where(flat_localmax[i_row])[0] for i_row in range(n_row)]
    
    # loop for trials
    for i_ap in range(n_ap_try):
        for i_row in range(i_row_center-1, 0-1, -1):
            this_d = np.abs(each_row_localmax[i_row]-ap_center_try[i_ap, i_row+1])
            if np.min(this_d) <= sigma:
                ap_center_try[i_ap, i_row] = each_row_localmax[i_row][np.argmin(this_d)]
            else:
                break
        for i_row in range(i_row_center+1, n_row, 1):
            this_d = np.abs(each_row_localmax[i_row]-ap_center_try[i_ap, i_row-1])
            if np.min(this_d) <= sigma:
                ap_center_try[i_ap, i_row] = each_row_localmax[i_row][np.argmin(this_d)]
            else:
                break
    ind_good_ap = np.sum(np.isfinite(ap_center_try), axis=1) == n_row
    n_ap = np.sum(ind_good_ap)
    ap_center = ap_center_try[ind_good_ap]
    
    return ap_center


#%%
figure()
plot()

# %%
figure();
plot(flat_smooth[1000]); 
#plot(np.diff(flat_smooth[1000]))
plot(flat_smooth[1000][1:]-flat_smooth[1000][:-1])

flat = fits.getdata("/hydrogen/projects/others/yujingcheng/20191212/quartz.fits.gz")[:2048,:1024]
star = fits.getdata("/hydrogen/projects/others/yujingcheng/20191212/hr4468.fits")[:2048,:1024]
star = fits.getdata("/media/cham/Ubuntu 17.1/b0111.fits")[:2048,:1024]

ap_center = trace_naive_max(flat, sigma=5, disperse_axis="col", )
print(ap_center.shape)
#imshow(np.log10(flat))
figure(); imshow(flat); plot(ap_center.T,np.arange(2048),"w")


ap = Aperture(ap_center.shape[0], ap_center, ap_width=10)
ap.get_image_info(flat)
ap.polyfit(4)


#%%

_blz, _norm = extract.make_normflat(flat, ap, ap_width=12)
figure(); imshow(_norm)
rextr = extract.extract_all(flat, ap, ap_width=12)


#%%

starbg = ap.background(star, q=(5,5))
rextr = extract.extract_all(star-starbg, ap, ap_width=10)

figure(); plot(star[1200]); plot(starbg[1200])
#%%
figure();
plot(np.arange(2048),rextr["spec_sum"][10].T / _blz[10].T)
plot(np.arange(2048)+2048,rextr["spec_sum"][11].T/ _blz[11].T)

#figure();plot(_blz.T)
#%%
star0  = fits.getdata("/media/cham/Ubuntu 17.1/obj4blue_multi.fits")
#%%

flux0 = star0[1,5,:].T
flux1 = rextr["spec_extr"][-5].T
flux2 = rextr["spec_extr2"][-5].T
wave = np.arange(2048)


flux0n,_ = normalize_spectrum_iter(wave, flux0,niter=3)
flux1n,_ = normalize_spectrum_iter(wave, flux1,niter=3)
flux2n,_ = normalize_spectrum_iter(wave, flux2,niter=3)

figure();
plot(flux0n);
plot(flux1n+1);
plot(flux2n+1);

import joblib
joblib.dump(rextr, "/media/cham/Ubuntu 17.1/rextr_songcn.dump")
#import joblib
#rextr = joblib.load("~")


# -*- coding: utf-8 -*-
"""
Created on Tue May 30 13:27:58 2017

@author: cham

@SONG: RMS =  0.00270124312246
delta_rv = 299792458/5500*0.00270124312246 = 147.23860278870518 m/s

LAMOST: R~1800,  299792.458/6000*3A = 150km/s WCALIB: 10km/s delta_rv = 5km/s
MMT: R~2500,  299792.458/R = 119.9km/s  rms=0.07A, delta_rv = 5km/s
SONG: 1800,  299792.458/6000*3A = 150km/s delta_rv = 5km/s
"""

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.pyplot import cm
from scipy.stats import binned_statistic

from twodspec import calibration as calib


def rms(x):
    return np.sqrt(np.mean(np.square(x)))


def nanrms(x):
    return np.sqrt(np.nanmean(np.square(x)))


def calibrate(thar1d_simple, thar_solution_temp, thar_list,
              poly_order=(5, 10), slit=0):
    wave_temp, thar_temp, order_temp = thar_solution_temp

    # 1.initial solution
    print("@SONG: [ThAr] 2D cross-correlation ...")
    shift, corr2d = calib.thar1d_corr2d(
        thar1d_simple, thar_temp, y_shiftmax=3, x_shiftmax=20)
    print("@SONG: [ThAr] the shift is ", shift)
    wave_init = calib.interpolate_wavelength(
        wave_temp, shift, thar_temp, thar1d_simple)
    order_init = calib.interpolate_order(
        order_temp, shift, thar1d_simple) + 80

    print("@SONG: [ThAr] refine wavelength ...""")
    # 2.fit Gaussians to Thar lines
    lc_coord, lc_order, lc_thar, popt, pcov = calib.refine_thar_positions(
        wave_init, order_init, thar1d_simple, thar_list,
        fit_width=.3, lc_tol=.1, k=3, n_jobs=-1, verbose=10)

    # select using center deviation & line SNR
    ind_good0 = np.logical_and(
        np.abs(popt[:, 2] - lc_thar) < 5,  # (popt[:,3]*3),
        (popt[:, 1] / np.sqrt(2. * np.pi) / popt[:, 3] / np.abs(
            popt[:, 0])) > .1)
    print(np.sum(ind_good0))

    # 3.rejections of outliers
    ind_good1 = calib.clean_thar_polyfit1d_reject(
        lc_coord, lc_order, lc_thar, popt, ind_good0=ind_good0, deg=1, w=None,
        epsilon=0.002, n_reserve=8)
    print(np.sum(ind_good1))

    # 4. fit final solution
    print("@SONG: [ThAr] fit wavelength solution ")
    x_mini_lsq, ind_good_thar, scaler_coord, scaler_order, scaler_ml = calib.fit_grating_equation(
        lc_coord, lc_order, lc_thar, popt, pcov, ind_good_thar0=ind_good1,
        poly_order=poly_order, max_dev_threshold=.003, n_iter=1, lar=False,
        nl_eachorder=5)

    # construct grids for coordinates & order
    grid_coord, grid_order = np.meshgrid(np.arange(thar1d_simple.shape[1]),
                                         np.arange(
                                             thar1d_simple.shape[0]) + 80)

    # 4'.fit grating function
    sgrid_fitted_wave = calib.grating_equation_predict(
        grid_coord, grid_order, x_mini_lsq, poly_order,
        scaler_coord, scaler_order, scaler_ml)

    # 5.get the fitted wavelength
    lc_thar_fitted = calib.grating_equation_predict(
        lc_coord, lc_order, x_mini_lsq, poly_order,
        scaler_coord, scaler_order, scaler_ml)

    results = [sgrid_fitted_wave, lc_thar_fitted]

    # 6.figures for diagnostics
    bins = np.arange(4500, 7500, 500)
    bins_med, _, _ = binned_statistic(lc_thar[ind_good_thar],
                                      lc_thar_fitted[ind_good_thar] - lc_thar[
                                          ind_good_thar], statistic=np.median,
                                      bins=bins)
    bins_rms, _, _ = binned_statistic(lc_thar[ind_good_thar],
                                      lc_thar_fitted[ind_good_thar] - lc_thar[
                                          ind_good_thar], statistic=nanrms,
                                      bins=bins)

    """ [Figure]: calibration diagnostics """
    fig = plt.figure(figsize=(12, 8))
    fig.add_subplot(111)
    plt.plot(lc_thar, lc_thar_fitted - lc_thar, '.')
    plt.plot(lc_thar[ind_good_thar],
             lc_thar_fitted[ind_good_thar] - lc_thar[ind_good_thar], 'r.')
    plt.errorbar(bins[:-1] + np.diff(bins) * .5, bins_med, bins_rms, color='k',
                 ecolor='k')
    plt.xlim(4300, 7200)
    plt.ylim(-.008, .008)
    plt.xlabel("Wavelength (A)")
    plt.ylabel("$\lambda(solution)-\lambda(true)$")
    plt.title("RMS = {:05f} A for SLIT {:d} [{} lines]".format(
        rms(lc_thar_fitted[ind_good_thar] - lc_thar[ind_good_thar]), slit,
        len(ind_good_thar)))
    plt.legend(["deviation of all lines", "deviation of used lines",
                "mean RMS in bins"])
    fig.tight_layout()
    # fig.savefig(dir_work+"thar{}_{:s}".format(slit, thar_fn.replace(".fits", "_diagnostics.svg")))
    # plt.close(fig)
    results.append(fig)

    fig = plt.figure(figsize=(24, 8))
    plt.imshow(np.log10(thar1d_simple), aspect='auto', cmap=cm.viridis,
               vmin=np.nanpercentile(np.log10(thar1d_simple), 5),
               vmax=np.nanpercentile(np.log10(thar1d_simple), 95))
    plt.plot(lc_coord, lc_order - 80, ls='', marker='s', mfc='None', mec='b')
    plt.plot(lc_coord[ind_good_thar], lc_order[ind_good_thar] - 80, ls='',
             marker='s', mfc='None', mec='r')
    plt.xlabel("CCD Coordinate")
    plt.ylabel("Order")
    plt.colorbar()
    fig.tight_layout()
    # fig.savefig(dir_work+"thar{}_{:s}".format(slit, thar_fn.replace(".fits", "_used_lines.svg")))
    # plt.close(fig)
    results.append(fig)

    return results
